import sys
import warnings

import numpy as np
from scipy.stats import rankdata
from scipy.linalg import cho_factor, cho_solve
from copy import deepcopy
from dataclasses import fields, MISSING


TAB = '    '
SP2 = '  '
maxfloat = np.float128 if hasattr(np, 'float128') else np.longdouble
_slsqp_epsilon = np.sqrt(np.finfo(float).eps)  # scipy's default value for the SLSQP epsilon parameter


class Struct:
    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}[attributes: ' + ', '.join([k for k in self.__dict__.keys()]) + ']'


class ReprMixin:
    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}\n' + '\n'.join([f'\t{k}: {v}' for k, v in self.__dict__.items()])


def create_struct_with_reprmixin(class_name):
    return type(class_name, (ReprMixin,), {})()


def print_class_instance(instance, attr_class_only=(), attr_replace_string=None):
    txt = f'{instance.__class__.__name__}'
    for k, v in instance.__dict__.items():
        if k in attr_class_only:
            txt += f"\n\t{k}: {v.__class__.__name__}"
        elif attr_replace_string is not None and k in attr_replace_string:
            txt += f"\n\t{k}: {attr_replace_string[k]}"
        elif isinstance(v, dict):
            txt += f"\n\t{k}: { {k_: np.array2string(np.array(v_), precision=4, threshold=20, separator=', ') for k_, v_ in v.items()} }"
        elif isinstance(v, list):
            if isinstance(v[0], dict):
                txt += f'\n\t{k}:[\n'
                for i in range(min(5, len(v))):
                    txt += '\t\t{' + ', '.join([f"'{k_}': {np.array2string(np.array(v_), precision=4, separator=', ')}" for k_, v_ in v[i].items()]) + '}\n'
                if len(v) > 10:
                    txt += '\t\t[...]\n'
                if len(v) > 5:
                    for i in range(max(-5, -len(v)), 0):
                        txt += '\t\t{' + ', '.join([f"'{k_}': {np.array2string(np.array(v_), precision=4, separator=', ')}" for k_, v_ in v[i].items()]) + '}\n'
                txt += '\t]'
            elif isinstance(v[0], float):
                txt += f"\n\t{k}: {np.array2string(np.array(v), precision=4, threshold=50, separator=', ')}"
        elif isinstance(v, np.ndarray):
            txt += f"\n\t{k}: {np.array2string(np.array(v), precision=4, threshold=50, separator=', ')}"
        else:
            txt += f"\n\t{k}: {v}"
    # print(txt)
    return txt

def reset_dataclass_on_init(cls):
    """
    For a @dataclass:
      - On each new instance, reset:
          * non-field class attributes
          * dataclass fields that have a direct default (not default_factory)
        back to the values they had at class definition time.
    """

    # ---- 1) Capture non-field class attributes (true class attrs) ----
    dataclass_field_names = set(cls.__dataclass_fields__.keys())

    original_class_attrs = {}
    for name, value in cls.__dict__.items():
        if name.startswith("__") and name.endswith("__"):
            continue
        if name in dataclass_field_names:
            # handled separately as dataclass fields
            continue
        if callable(value):
            continue

        original_class_attrs[name] = deepcopy(value)

    # ---- 2) Capture dataclass field defaults (for __init__ defaults + class attrs) ----
    dc_fields = list(fields(cls))

    # fields that appear as parameters with defaults in __init__
    fields_with_any_default = []
    field_default_kinds = []       # "value" or "factory"
    stored_field_defaults = []     # original defaults or factories

    for f in dc_fields:
        if f.default is not MISSING or f.default_factory is not MISSING:
            fields_with_any_default.append(f)
            if f.default_factory is not MISSING:
                field_default_kinds.append("factory")
                stored_field_defaults.append(f.default_factory)
            else:
                field_default_kinds.append("value")
                stored_field_defaults.append(deepcopy(f.default))

    # For fields with a *value* default, there is usually a class attribute with that name.
    # We'll reset that class attribute too.
    field_class_default_indices = {}
    for idx, (f, kind, stored) in enumerate(
        zip(fields_with_any_default, field_default_kinds, stored_field_defaults)
    ):
        if kind == "value" and hasattr(cls, f.name):
            field_class_default_indices[f.name] = idx

    # ---- 3) Wrap the original __init__ ----
    original_init = cls.__init__

    # Sanity: number of __init__ defaults should match number of fields with defaults
    orig_defaults_len = len(original_init.__defaults__ or ())
    if orig_defaults_len != len(fields_with_any_default):
        # You *can* turn this into an assert if you prefer it to fail loudly
        pass

    def __init__(self, *args, **kwargs):
        # 3a) reset non-field class attributes
        for name, value in original_class_attrs.items():
            setattr(cls, name, deepcopy(value))

        # 3b) reset class attributes for dataclass fields with direct defaults
        for fname, idx in field_class_default_indices.items():
            setattr(cls, fname, deepcopy(stored_field_defaults[idx]))

        # 3c) rebuild the original __init__ defaults for this call
        if fields_with_any_default:
            new_defaults = []
            for kind, stored in zip(field_default_kinds, stored_field_defaults):
                if kind == "factory":
                    # default_factory fields keep their original factory
                    new_defaults.append(stored)
                else:
                    # direct defaults get a fresh deep copy
                    new_defaults.append(deepcopy(stored))
            original_init.__defaults__ = tuple(new_defaults)

        # Call the real dataclass-generated __init__
        original_init(self, *args, **kwargs)

    cls.__init__ = __init__
    return cls


def listlike(x):
    return np.array(x).ndim > 0


def empty_list(n, *shapes_or_None):
    if (len(shapes_or_None) == 1) and shapes_or_None[0] is None:
        return [None] * n
    else:
        return [np.empty([shape[i] if listlike(shape) else shape for shape in shapes_or_None]) for i in range(n)]


def fmp(v, k=3):
    # 1. Format to exactly k decimal places
    s = f"{v:.{k}f}"

    # 2. If it ends in k zeros (e.g. .000 for k=3), strip it all (for cases like 1.000 -> 1)
    if s.endswith('.' + '0'*k):
        return s[:-(k+1)]

    # 1.110 -> 1.11
    orig_str = str(v)
    if '.' in orig_str:
        decimal_places = len(orig_str.split('.')[1])
        # If original has <k decimals, keep it as is.
        # If original has k+ decimals, truncate to k.
        precision = min(decimal_places, k)
        return f"{v:.{precision}f}"

    return str(int(v))


def _check_param(x):
    if hasattr(x, '__len__'):
        if len(x) == 2:
            return x
        elif len(x) == 1:
            return [x[0], x[0]]
        else:
            print(f'Something went wrong, parameter array has length {len(x)}')
    else:
        return [x, x]


def _check_criteria(x):
    if hasattr(x[0], '__len__'):
        return x
    else:
        return [x, x]


def pearson2d(x, y, axis=-1):
    assert axis in (-2, -1, 0, 1), 'axis must be one of -2, -1, 0, 1'
    x, y = np.asarray(x), np.asarray(y)
    mx, my = np.nanmean(x, axis=axis), np.nanmean(y, axis=axis)
    if axis in (0, -2):
        xm, ym = x - mx, y - my
    else:
        xm, ym = x - mx[..., None], y - my[..., None]
    r_num = np.nansum(xm * ym, axis=axis)
    r_den = np.sqrt(np.nansum(xm ** 2, axis=axis) * np.nansum(ym ** 2, axis=axis))
    r_den[np.isclose(r_den, 0)] = np.nan
    r = r_num / r_den
    return r


def spearman2d(x, y, axis=-1):
    assert axis in (-2, -1, 0, 1), 'axis must be one of -2, -1, 0, 1'
    x, y = np.asarray(x), np.asarray(y)
    xr, yr = rankdata(x, axis=axis), rankdata(y, axis=axis)
    mxr, myr = np.nanmean(xr, axis=axis), np.nanmean(yr, axis=axis)
    if axis in (0, -2):
        xmr, ymr = xr - mxr, yr - myr
    else:
        xmr, ymr = xr - mxr[..., None], yr - myr[..., None]
    r_num = np.nansum(xmr * ymr, axis=axis)
    r_den = np.sqrt(np.nansum(xmr ** 2, axis=axis) * np.nansum(ymr ** 2, axis=axis))
    r_den[np.isclose(r_den, 0)] = np.nan
    r = r_num / r_den
    return r


# def cov_from_hessian(H, symmetrize=True):
#     Hs = 0.5 * (H + H.T) + 1e-12 * np.eye(len(H)) if symmetrize else H + 1e-12 * np.eye(len(H))
#     c, lower = cho_factor(Hs, lower=True, check_finite=False)  # raises LinAlgError if not SPD
#     cov = cho_solve((c, lower), np.eye(Hs.shape[0]), check_finite=False)
#     # numerical cleanup: enforce symmetry
#     cov = 0.5 * (cov + cov.T)
#     return cov

def cov_from_hessian(H, symmetrize=True):
    Hs = 0.5 * (H + H.T) if symmetrize else np.asarray(H, dtype=float)
    Hs = Hs + 1e-12 * np.eye(Hs.shape[0])

    try:
        c, lower = cho_factor(Hs, lower=True, check_finite=False)
        cov = cho_solve((c, lower), np.eye(Hs.shape[0]), check_finite=False)
    except np.linalg.LinAlgError:
        from remeta.fit import ridge_hessian
        Hr = ridge_hessian(Hs)
        if Hr is None:
            return None
        c, lower = cho_factor(Hr, lower=True, check_finite=False)
        cov = cho_solve((c, lower), np.eye(Hr.shape[0]), check_finite=False)

    # numerical cleanup: enforce symmetry
    cov = 0.5 * (cov + cov.T)

    return cov


def se_from_cov(cov):
    var = np.diag(cov)
    se = np.sqrt(var)
    return se


def compute_cov_criteria(cov_full, idx_crit):
    cov_gaps = cov_full[np.ix_(idx_crit, idx_crit)]
    J = np.tril(np.ones((len(idx_crit), len(idx_crit))))
    cov_crit = J @ cov_gaps @ J.T
    return cov_crit


def compute_criterion_bias(criteria, cov_crit):
    """Idea: compute criterion bias as a weighted sum of differences from Bayes-optiomal criteria.
       The weights are the uncertainty estimates (SEs) of the criteria; we slightly improve on this
       by also considering the correlation structure between criterion uncertainties, i.e. the full covariance
       matrix.
    """

    k = len(criteria) + 1
    crit_bayes = np.arange(1/k, 1-1e-10, 1/k)
    diff = criteria - crit_bayes
    one = np.ones_like(criteria)

    c, lower = cho_factor(cov_crit)
    # numer = (multivariate) weighted sum of criterion differences
    numer = one @ cho_solve((c, lower), diff)
    # denom = (multivariate) sum of all weights
    denom = one @ cho_solve((c, lower), one)
    bias_crit = numer / denom
    bias_crit_se = np.sqrt(1 / denom)

    return bias_crit, bias_crit_se

def compute_choice_bias(stimuli, choices, smooth=0.5):
    # levels = np.sort(np.unique(np.abs(stimuli)))
    # ntrials = len(stimuli)

    # Unique signed levels
    x_levels = np.unique(stimuli)
    magnitudes = np.unique(np.abs(x_levels))
    magnitudes = magnitudes[magnitudes > 0]

    # Aggregate counts
    k = {}
    n = {}
    for x in x_levels:
        mask = (stimuli == x)
        n[x] = mask.sum()
        k[x] = choices[mask].sum()

    D_list, w_list = [], []

    for m in magnitudes:
        if m in k and -m in k:
            # empirical proportions
            p_plus = k[m] / n[m]
            p_minus = k[-m] / n[-m]
            # smoothed proportions for variance
            p_plus_var = (k[m] + smooth) / (n[m] + 2*smooth)
            p_minus_var = (k[-m] + smooth) / (n[-m] + 2*smooth)
            var_plus = p_plus_var * (1 - p_plus_var) / n[m]
            var_minus = p_minus_var * (1 - p_minus_var) / n[-m]
            D = p_plus + p_minus - 1
            var_D = var_plus + var_minus
            if var_D <= 0:
                continue
            w = 1.0 / var_D
            D_list.append(D)
            w_list.append(w)

    D_arr = np.array(D_list)
    w_arr = np.array(w_list)

    bias_hat = np.sum(w_arr * D_arr) / np.sum(w_arr)

    return bias_hat


def _pav_isotonic(y, w=None):
    """
    Pool-Adjacent-Violators (PAV) for isotonic regression (nondecreasing).
    Returns fitted values yhat with same length as y.
    """
    y = np.asarray(y, dtype=float)
    n = y.size
    if w is None:
        w = np.ones(n, dtype=float)
    else:
        w = np.asarray(w, dtype=float)
        if w.shape != y.shape:
            raise ValueError("w must have same shape as y")

    # Each point starts as its own block
    v = y.copy()
    ww = w.copy()
    start = np.arange(n)
    end = np.arange(n)

    # Stack of block indices
    m = 0
    for i in range(n):
        start[m] = i
        end[m] = i
        v[m] = y[i]
        ww[m] = w[i]
        m += 1

        # Merge while monotonicity violated
        while m >= 2 and v[m-2] > v[m-1]:
            new_w = ww[m-2] + ww[m-1]
            new_v = (ww[m-2]*v[m-2] + ww[m-1]*v[m-1]) / new_w
            ww[m-2] = new_w
            v[m-2] = new_v
            end[m-2] = end[m-1]
            m -= 1

    # Expand blocks
    yhat = np.empty(n, dtype=float)
    for b in range(m):
        yhat[start[b]:end[b]+1] = v[b]
    return yhat


def compute_choice_bias_horizontal_old(stimuli, choices, smooth=0.5):
    """
    Horizontal (stimulus-units) bias estimator using:
      - symmetry deviations across matched ±m levels
      - inverse-variance weighting from binomial sampling
      - nonparametric slope estimates from isotonic-smoothed proportions

    Parameters
    ----------
    stimuli : (T,) array-like of floats
        Signed stimulus intensities.
    choices : (T,) array-like of ints
        Binary choices: 0=negative category, 1=positive category.
    smooth : float
        Jeffreys-style smoothing for variance stability in Var(p).

    Returns
    -------
    mu_hat : float
        Estimated horizontal bias (PSE shift) in stimulus units.
        Positive mu_hat means the curve is shifted right (harder to choose 1).
        (Sign convention depends on defining p(x)=P(choice=1|x).)
    """

    # Aggregate by unique signed intensity
    x_levels = np.unique(stimuli)
    x_levels.sort()
    n = np.zeros_like(x_levels, dtype=int)
    k = np.zeros_like(x_levels, dtype=int)

    for i, x in enumerate(x_levels):
        mask = (stimuli == x)
        n[i] = int(mask.sum())
        k[i] = int(choices[mask].sum())

    p_hat = k / n

    # Smoothed p for variance stability
    p_var = (k + smooth) / (n + 2.0 * smooth)
    var_p = p_var * (1.0 - p_var) / n

    # Monotone smoothing (optional but very helpful) via isotonic regression
    # Weighted by n to respect unequal trial counts
    p_iso = _pav_isotonic(p_hat, w=n.astype(float))

    # Slope estimate at each x_level using finite differences on isotonic fit
    slope = np.empty_like(p_iso, dtype=float)
    if x_levels.size == 1:
        raise ValueError("Need more than one stimulus level to estimate slope.")

    # One-sided at edges, central inside
    slope[0] = (p_iso[1] - p_iso[0]) / (x_levels[1] - x_levels[0])
    slope[-1] = (p_iso[-1] - p_iso[-2]) / (x_levels[-1] - x_levels[-2])
    for i in range(1, x_levels.size - 1):
        slope[i] = (p_iso[i+1] - p_iso[i-1]) / (x_levels[i+1] - x_levels[i-1])

    # Build matched ±m pairs
    # Map x -> index for quick lookup
    idx = {float(x): i for i, x in enumerate(x_levels)}

    mags = np.unique(np.abs(x_levels))
    mags = mags[mags > 0]  # exclude 0

    D_list, s_list, varD_list, w_list, m_list = [], [], [], [], []

    for m in mags:
        mp = float(m)
        mm = float(-m)
        if mp in idx and mm in idx:
            ip = idx[mp]
            im = idx[mm]

            # Symmetry deviation uses empirical p_hat (could also use p_iso; choose empiric here)
            Dm = p_hat[ip] + p_hat[im] - 1.0

            # Sampling variance for Dm from (smoothed) binomial variance
            varDm = var_p[ip] + var_p[im]
            if varDm <= 0:
                continue

            # Slope estimate at ±m from isotonic fit; average them
            sm = 0.5 * (slope[ip] + slope[im])

            # If slope is ~0, this magnitude carries little info about a horizontal shift
            # (we can safely drop it)
            if sm == 0 or not np.isfinite(sm):
                continue

            wm = 1.0 / varDm

            D_list.append(Dm)
            s_list.append(sm)
            w_list.append(wm)

    D = np.asarray(D_list, dtype=float)
    s = np.asarray(s_list, dtype=float)
    w = np.asarray(w_list, dtype=float)

    # Weighted LS solution for Dm ≈ -2 s_m mu
    num = np.sum(w * s * D)
    den = np.sum(w * s * s)

    mu_hat = num / (2.0 * den)

    return float(mu_hat)



def compute_choice_bias_horizontal(stimuli, choices, smooth=0.5, eps=1e-12):
    """
    Stimulus-space choice bias estimated nonparametrically using ALL stimulus levels with inverse-variance weights.

    Parameters
    ----------
    stimuli : (T,) array-like
        Signed stimulus intensities (numeric).
    choices : (T,) array-like
        Binary choices: 0 = negative category, 1 = positive category.
    smooth : float, default 0.5
        Jeffreys-style smoothing for variance stability (used in Var estimates only).
    eps : float, default 1e-12
        Floors tiny variances to avoid infinite weights.

    Returns
    -------
    bias_hat : float
        Estimated bias in stimulus units (PSE = x where fitted p(x)=0.5).
        If your logistic is p(x)=sigmoid(beta*(x - mu)), then bias_hat ≈ mu.
        Returns +inf/-inf if 0.5 is not bracketed by the fitted curve.
    """
    stimuli = np.asarray(stimuli, dtype=float)
    choices = np.asarray(choices, dtype=int)

    if stimuli.ndim != 1 or choices.ndim != 1:
        raise ValueError("stimuli and choices must be 1D arrays.")
    if stimuli.size != choices.size:
        raise ValueError("stimuli and choices must have the same length.")
    if not np.all((choices == 0) | (choices == 1)):
        raise ValueError("choices must be binary (0/1).")

    # Aggregate by unique intensity levels
    x_levels, inv = np.unique(stimuli, return_inverse=True)
    order = np.argsort(x_levels)
    x_levels = x_levels[order]
    inv = order[inv]  # remap to sorted order indices

    n = np.bincount(inv)
    k = np.bincount(inv, weights=choices).astype(float)
    if x_levels.size < 2:
        raise ValueError("Need at least two distinct stimulus levels.")

    p_hat = k / n

    # Sampling variance (binomial), with smoothing to avoid var=0 at p=0 or 1
    p_var = (k + smooth) / (n + 2.0 * smooth)
    var_p = p_var * (1.0 - p_var) / n
    w = 1.0 / np.maximum(var_p, eps)  # inverse-variance weights

    # Nonparametric monotone psychometric using IV weights (uses ALL levels)
    p_iso = _pav_isotonic(p_hat, w=w)

    # If 0.5 not bracketed, PSE lies outside tested range
    if np.all(p_iso < 0.5):
        return float(np.inf)
    if np.all(p_iso > 0.5):
        return float(-np.inf)

    # If there's a plateau exactly at 0.5, return its midpoint in x
    eq = (p_iso == 0.5)
    if np.any(eq):
        xs = x_levels[eq]
        return float(0.5 * (xs.min() + xs.max()))

    # Find the crossing interval (monotone => should be unique)
    # We find the first index i where p_iso[i] >= 0.5, then interpolate with i-1.
    i_hi = int(np.argmax(p_iso >= 0.5))
    i_lo = i_hi - 1
    if i_lo < 0:
        return float(-np.inf)

    x0, x1 = x_levels[i_lo], x_levels[i_hi]
    p0, p1 = p_iso[i_lo], p_iso[i_hi]
    if p1 == p0:
        return float(0.5 * (x0 + x1))

    t = (0.5 - p0) / (p1 - p0)
    bias = float(x0 + t * (x1 - x0))
    return -bias



def print_warnings(w):
    if len(w):
        print('\tWarnings that occured during model estimation:')
    for el in set([w_.message.args[0] for w_ in w]):
        if 'delta_grad == 0.0' not in el:
            print('\t\tWarning: ' + el)


def raise_warning_in_catch_block(msg, category, w):
    warnings.warn(msg, category=category)
    if len(w):
        sys.stderr.write(warnings.formatwarning(
            w[-1].message, w[-1].category, w[-1].filename, w[-1].lineno, line=w[-1].line
        ))

# def type2roc(correct, conf, nbins=5):
#     # Calculate area under type 2 ROC
#     #
#     # correct - vector of 1 x ntrials, 0 for error, 1 for correct
#     # conf - vector of continuous confidence ratings between 0 and 1
#     # nbins - how many bins to use for discretization
#
#     bs = 1 / nbins
#     h2, fa2 = np.full(nbins, np.nan), np.full(nbins, np.nan)
#     for c in range(nbins):
#         if c:
#             h2[nbins - c - 1] = np.sum((conf > c*bs) & (conf <= (c+1)*bs) & correct.astype(bool)) + 0.5
#             fa2[nbins - c - 1] = np.sum((conf > c*bs) & (conf <= (c+1)*bs) & ~correct.astype(bool)) + 0.5
#         else:
#             h2[nbins - c - 1] = np.sum((conf >= c * bs) & (conf <= (c + 1) * bs) & correct.astype(bool)) + 0.5
#             fa2[nbins - c - 1] = np.sum((conf >= c * bs) & (conf <= (c + 1) * bs) & ~correct.astype(bool)) + 0.5
#
#     h2 /= np.sum(h2)
#     fa2 /= np.sum(fa2)
#     cum_h2 = np.hstack((0, np.cumsum(h2)))
#     cum_fa2 = np.hstack((0, np.cumsum(fa2)))
#
#     k = np.full(nbins, np.nan)
#     for c in range(nbins):
#         k[c] = (cum_h2[c+1] - cum_fa2[c])**2 - (cum_h2[c] - cum_fa2[c+1])**2
#
#     auroc2 = 0.5 + 0.25*np.sum(k)
#
#     return auroc2

def discretize_confidence_with_bounds(x, bounds):
    confidence = np.full(x.shape, np.nan)
    bounds = np.hstack((bounds, np.inf))
    for i, b in enumerate(bounds[:-1]):
        confidence[(bounds[i] <= x) & (x < bounds[i + 1])] = i + 1
    return confidence

def print_dataset_characteristics(sim):
    print('----------------------------------')
    if sim.cfg.skip_type2:
        print('..Generative parameters:')
        print(f'{TAB}Type 1 noise distribution: {sim.cfg.param_type1_noise.model}')
        for p, v in sim.params_type1.items():
            print(f'{TAB}{p}: {np.array2string(np.array(v), precision=3)}')
    else:
        print('..Generative model:')
        print(f'{TAB}Type 1 noise distribution: {sim.cfg.param_type1_noise.model}')
        print(f'{TAB}Type 2 noise type: {sim.cfg.type2_noise_type}')
        print(f'{TAB}Type 2 noise distribution: {sim.cfg.param_type2_noise.model}')
        print('..Generative parameters:')
        for p, v in sim.params.items():
            print(f'{TAB}{p}: {np.array2string(np.array(v), precision=3)}')
        if sim.params_extra is not None:
            if 'type2_criteria_bias' in sim.params_extra:
                print(f"{TAB}{TAB}[extra] Criterion bias: {sim.params_extra['type2_criteria_bias']:.4f}")
            if 'type2_criteria_confidence_bias' in sim.params_extra:
                print(f"{TAB}{TAB}[extra] Criterion-based confidence bias: {sim.params_extra['type2_criteria_confidence_bias']:.4f}")
            # if 'type2_criteria_absdev' in sim.params_extra:
            #     print(f"{TAB}{TAB}[extra] Criterion absolute deviation: {sim.params_extra['type2_criteria_absdev']:.4f}")
    print('..Descriptive statistics:')
    print(f'{TAB}No. subjects: {sim.nsubjects}')
    print(f"{TAB}No. samples: {np.array2string(np.array(sim.nsamples).squeeze(), separator=', ', threshold=3)}")
    if sim.type1_stats is not None:
        print(f"{TAB}Accuracy: {100 * sim.type1_stats['accuracy']:.1f}% correct")
        print(f"{TAB}d': {sim.type1_stats['dprime']:.1f}")
        # print(f"{TAB}Choice bias: {('-', '+')[int(sim.type1_stats['choice_bias'] > 0.5)]}{100*np.abs(sim.type1_stats['choice_bias'] - 0.5):.1f}%")
        print(f"{TAB}Choice bias: {100*sim.type1_stats['choice_bias']:.1f}%")
    if not sim.cfg.skip_type2 and sim.type2_stats is not None:
        print(f"{TAB}Confidence: {sim.type2_stats['confidence']:.2f}")
        print(f"{TAB}M-Ratio: {sim.type2_stats['mratio']:.2f}")
        print(f"{TAB}AUROC2: {sim.type2_stats['auroc2']:.2f}")
    print('----------------------------------')



import os
import pathlib
import pickle
import timeit
import warnings

try:  # only necessary if multiple cores should be used
    from multiprocessing_on_dill.pool import Pool as DillPool
except ModuleNotFoundError:
    pass

import numpy as np
from scipy.optimize import OptimizeResult
from scipy.special import expit, ndtr, erfinv

from .configuration import Configuration
from .type2_dist import get_type2_dist
from .fit import subject_estimation, group_estimation
from .simulation import simulate
from .modelspec import ModelData, Data, Summary
from .plotting import plot_psychometric, plot_stimulus_versus_confidence, plot_confidence_histogram
from .transform import (compute_signal_dependent_type1_noise, logistic, type1_evidence_to_confidence,
                        confidence_to_type1_evidence, confidence_to_type1_noise, type1_noise_to_confidence,
                        compute_nonlinear_encoding)
from .util import _check_param, TAB, SP2, print_dataset_characteristics, print_warnings, empty_list

class ReMeta:

    def __init__(
        self,
        cfg: Configuration = None,
        **kwargs
    ):
        """

        Usage:
            ```
            rem = remeta.ReMeta()
            ```

            for a default model, or otherwise:

            ```
            cfg = remeta.Configuration()
            cfg.<some_setting> = <some_value>
            rem = remeta.ReMeta(cfg)
            ```

        Args:
            cfg: Configuration object. If None is passed, the default configuration (but see kwargs).
            **kwargs: kwargs entries are passed to the Configuration

        """

        if cfg is None:
            # Set configuration attributes that match keyword arguments
            cfg_kwargs = {k: v for k, v in kwargs.items() if k in Configuration.__dict__}
            self.cfg = Configuration(**cfg_kwargs)
        else:
            self.cfg = cfg
            for k, v in kwargs.items():
                if k in Configuration.__dict__:
                    setattr(self.cfg, k, v)
        self.cfg.setup()

        self.modeldata = ModelData(cfg=self.cfg)
        self.data = None
        self.result = None

        self.type1_is_fitted = False
        self.type2_is_fitted = False


    def fit(
        self,
        stimuli: list[float] | list[list[float]]  | np.typing.NDArray[float],
        choices: list[float] | list[list[float]]  | np.typing.NDArray[float],
        confidence: list[float] | list[list[float]]  | np.typing.NDArray[float] = None,
        verbosity: int = 1,
        silence_warnings: bool = False
    ):
        """

        Usage:
            ```
            rem = ReMeta()
            rem.fit(stimuli, choices, confidence)
            ```

        Args:
            stimuli: 1d or 2d array or list of signed stimulus intensities
            choices: 1d or 2d array or list of choices (coded as -1/1 or 0/1)
            confidence: 1d or 2d array or list of confidence ratings (normalized to the range 0-1)
            verbosity: verbosity level (possible values: 0, 1, 2)
            silence_warnings: if `True`, warnings during model fitting are supressed.
        """

        self.data = Data(self.cfg, stimuli, choices, confidence)

        self.result = Summary(self.data, self.cfg)

        self.fit_type1(verbosity=verbosity, store_final_results=self.cfg.skip_type2, silence_warnings=silence_warnings)

        if not self.cfg.skip_type2:
            self.fit_type2(verbosity=verbosity, silence_warnings=silence_warnings)

        return self


    def fit_type1(
        self,
        stimuli: None | list[float] | list[list[float]]  | np.typing.NDArray[float] = None,
        choices: None | list[float] | list[list[float]]  | np.typing.NDArray[float] = None,
        confidence: None | list[float] | list[list[float]]  | np.typing.NDArray[float] = None,
        store_final_results: bool = True,
        verbosity: int = 1,
        silence_warnings: bool = False
    ):
        """

        Usage:
            ```
            rem = ReMeta()
            rem.fit_type1(stimuli, choices)
            ```

        Args:
            stimuli: 1d or 2d array or list of signed stimulus intensities
            choices: 1d or 2d array or list of choices (coded as -1/1 or 0/1)
            confidence: 1d or 2d array or list of confidence ratings (normalized to the range 0-1)
            store_final_results: if `True`, save final results. Mostly used internally - will be set to `False`,
                if followed by `fit_type2`.
            verbosity: verbosity level (possible values: 0, 1, 2)
            silence_warnings: if `True`, warnings during model fitting are supressed.
        """

        if self.data is None:
            if stimuli is None or choices is None:
                raise ValueError('If the data attribute of the ReMeta instance is None, at least stimuli '
                                 'and choices have to be passed to fits_type1_subject()')
            else:
                self.data = Data(self.cfg, stimuli, choices, confidence)

        if verbosity >= 1:
            print(f'Dataset characteristics:')
            print(f'{TAB}No. subjects: {self.data.nsubjects}')
            print(f"{TAB}No. samples: {np.array2string(np.array(self.data.nsamples).squeeze(), separator=', ', threshold=3)}")
            print(f'{TAB}Accuracy: {100*self.data.stats.accuracy:.1f}% correct')
            print(f"{TAB}d': {self.data.stats.dprime:.3f}")
            print(f"{TAB}Choice bias: {100*self.data.stats.choice_bias:.1f}%")
            if not self.cfg.skip_type2 and not store_final_results:
                print(f"{TAB}Mean confidence: {self.data.stats.mean_confidence:.3f} "
                      f"(min: {np.min(self.data.c_conf):.3f}, max: {np.max(self.data.c_conf):.3f})")

        self.result = Summary(self.data, self.cfg)

        if verbosity:
            print('\n+++ Type 1 level +++')
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning, module='scipy.optimize',
                                    message='delta_grad == 0.0. Check if the approximated function is linear. If the '
                                            'function is linear better results can be obtained by defining the Hessian '
                                            'as zero instead of using quasi-Newton approximations.')

            fits_type1_subject, fit_type1_group = None, None
            if self.cfg._paramset_type1.nparams > 0:

                if verbosity:
                    print(f'{SP2}Subject-level estimation (MLE)')
                    tind = timeit.default_timer()

                # Single-subject fits via MLE
                use_multiproc_for_subject_loop = (self.cfg._optim_num_cores >= 8) and (self.data.nsubjects >= 8)
                def subject_loop(s):
                    return subject_estimation(
                        self.compute_type1_negll, self.cfg._paramset_type1, args=[s],
                        gridsearch=self.cfg.optim_type1_gridsearch,
                        scipy_solvers=self.cfg.optim_type1_scipy_solvers,
                        num_cores=1 if use_multiproc_for_subject_loop else self.cfg._optim_num_cores,
                        minimize_along_grid=self.cfg.optim_type1_minimize_along_grid,
                        global_minimization=self.cfg.optim_type1_global_minimization,
                        # fine_gridsearch=self.cfg.optim_type1_fine_gridsearch,
                        verbosity=verbosity, silence_warnings=silence_warnings
                    )
                if use_multiproc_for_subject_loop:
                    with DillPool(self.cfg._optim_num_cores) as pool:
                        fits_type1_subject = pool.map(subject_loop, range(self.data.nsubjects))
                else:
                    fits_type1_subject = [None for _ in range(self.data.nsubjects)]
                    for s in range(self.data.nsubjects):
                        if (verbosity > 0) and (self.data.nsubjects > 1):
                            print(f'{TAB} Subject {s + 1} / {self.data.nsubjects}')
                        fits_type1_subject[s] = subject_loop(s)
                # Store single-subject results
                params_subject = [fits_type1_subject[s].x for s in range(self.data.nsubjects)]
                params_hessian_subject = [fits_type1_subject[s].hessian for s in range(self.data.nsubjects)]
                self.result.type1.subject.store(
                    'type1', self.cfg, self.data, self.compute_type1_negll, params_subject,
                    hessian=params_hessian_subject, fit=fits_type1_subject,
                    execution_time=np.sum([fits_type1_subject[s].execution_time for s in range(self.data.nsubjects)])
                )

                if verbosity:
                    print(f'{TAB}.. finished ({timeit.default_timer() - tind:.1f} secs).')

                if self.data.nsubjects > 1:
                    idx_fe = np.array([i for i, p in enumerate(self.cfg._paramset_type1.parameters_flat.values()) if p.group == 'fixed'])
                    idx_re = np.array([i for i, p in enumerate(self.cfg._paramset_type1.parameters_flat.values()) if p.group == 'random'])
                    if (len(idx_fe) > 0) or (len(idx_re) > 0):

                        fit_type1_group = group_estimation(
                            fun=self.compute_type1_negll,
                            nsubjects=self.data.nsubjects,
                            params_init=params_subject,
                            bounds=self.cfg._paramset_type1.bounds,
                            idx_fe=idx_fe,
                            idx_re=idx_re,
                            num_cores=self.cfg._optim_num_cores,
                            max_iter=30, sigma_floor=1e-3,
                            verbosity=verbosity,
                            # tau=0.05
                        )

                        self.result.type1.init_group()
                        self.result.type1.group.store(
                            'type1', self.cfg, self.data, self.compute_type1_negll,
                            params=[fit_type1_group.x[s] for s in range(self.data.nsubjects)],
                            params_se=[fit_type1_group.x_se[s] for s in range(self.data.nsubjects)],
                            params_cov=[fit_type1_group.x_cov[s] for s in range(self.data.nsubjects)],
                            pop_mean_sd=fit_type1_group.x_re_pop_mean_sd,
                            execution_time=fit_type1_group.execution_time
                        )

            self.result.type1.store(cfg=self.cfg, data=self.data, fun=self.compute_type1_negll)
            if verbosity:
                self.result.type1.report_fit(self.cfg)
            self.type1_is_fitted = True

        if store_final_results:
            self.result.store(store_type1_only=True)

        if verbosity:
            print('Type 1 level finished')

        return self


    def fit_type2(self, verbosity=1, silence_warnings=False):

        # compute decision values
        self._compute_decision_values()

        if self.cfg.type2_noise_type == 'temperature':
            # We precompute a few things for noisy temperature model
            if self.cfg.param_type1_noise.model == 'normal':
                # normal: c = 2*ɸ(z) - 1 -> z = ɸ^-1(0.5*(c+1)) -> Jacobian = dz/dc = sqrt(π/2)*exp((erf^-1(c))**2)
                self.modeldata.precomputed.jacobian_temperature = \
                [np.sqrt(np.pi / 2.0) * np.exp(erfinv(np.minimum(1-1e-8, self.data.c_conf[s]))**2) for s in range(self.data.nsubjects)]
            elif self.cfg.param_type1_noise.model == 'logistic':
                # logistic: c = tanh(z * (π/2√3)) -> z = (2√3/π) * atanh(c) -> Jacobian = dz/dc = (2√3/π) / (1 - c**2)
                self.modeldata.precomputed.jacobian_temperature = \
                    [((2 * np.sqrt(3)) / np.pi) / (1 - np.minimum(1-1e-8, self.data.c_conf[s])**2) for s in range(self.data.nsubjects)]
            self.modeldata.precomputed.quintiles_temperature = np.arange(self.cfg.temperature_marg_res, 1, self.cfg.temperature_marg_res)

        if verbosity:
            print('\n+++ Type 2 level +++')

        with warnings.catch_warnings():  # noqa
            warnings.filterwarnings('ignore', module='scipy.optimize')
            if self.cfg._paramset_type2.nparams > 0:

                if verbosity:
                    print(f'{SP2}Subject-level estimation (MLE)')
                    tind = timeit.default_timer()
                    # print(f'{SP2}Scipy solvers: {self.cfg.optim_type1_scipy_solvers}')

                # Single-subject fits via MLE
                use_multiproc_for_subject_loop = (self.cfg._optim_num_cores >= 8) and (self.data.nsubjects >= 8)
                def subject_loop(s):
                    return subject_estimation(
                        self.compute_type2_negll, self.cfg._paramset_type2, args=[s],
                        gridsearch=self.cfg.optim_type2_gridsearch,
                        num_cores=1 if use_multiproc_for_subject_loop else self.cfg._optim_num_cores,
                        minimize_along_grid=self.cfg.optim_type2_minimize_along_grid,
                        global_minimization=self.cfg.optim_type2_global_minimization,
                        # fine_gridsearch=self.cfg.optim_type2_fine_gridsearch,
                        scipy_solvers=self.cfg.optim_type2_scipy_solvers, slsqp_epsilon=self.cfg.optim_type2_slsqp_epsilon,
                        verbosity=verbosity, silence_warnings=silence_warnings
                    )
                if use_multiproc_for_subject_loop:
                    with DillPool(self.cfg._optim_multiproc_cores_effective) as pool:
                        fits_type2_subject = pool.map(subject_loop, range(self.data.nsubjects))
                else:
                    fits_type2_subject = [None for _ in range(self.data.nsubjects)]
                    for s in range(self.data.nsubjects):
                        if (verbosity > 0) and (self.data.nsubjects > 1):
                            print(f'{TAB} Subject {s + 1} / {self.data.nsubjects}')
                        fits_type2_subject[s] = subject_loop(s)

                # Store single-subject results
                params_subject = [fits_type2_subject[s].x for s in range(self.data.nsubjects)]
                params_hessian_subject = [fits_type2_subject[s].hessian for s in range(self.data.nsubjects)]
                self.result.type2.subject.store(
                    'type2', self.cfg, self.data, self.compute_type2_negll, params_subject,
                    hessian=params_hessian_subject,
                    fit=fits_type2_subject,
                    execution_time=np.sum([fits_type2_subject[s].execution_time for s in range(self.data.nsubjects)])
                )

                if verbosity:
                    print(f'{TAB}.. finished ({timeit.default_timer() - tind:.1f} secs).')

                # Group fit
                if self.data.nsubjects > 1:
                    idx_fe = np.array([i for i, p in enumerate(self.cfg._paramset_type2.parameters_flat.values()) if p.group == 'fixed'])
                    idx_re = np.array([i for i, p in enumerate(self.cfg._paramset_type2.parameters_flat.values()) if p.group == 'random'])
                    if (len(idx_fe) > 0) or (len(idx_re) > 0):

                        fit_type2_group = group_estimation(
                            fun=self.compute_type2_negll,
                            nsubjects=self.data.nsubjects,
                            params_init=params_subject,
                            bounds=self.cfg._paramset_type2.bounds,
                            idx_fe=idx_fe,
                            idx_re=idx_re,
                            num_cores=self.cfg._optim_num_cores,
                            max_iter=30, sigma_floor=1e-3,
                            verbosity=verbosity
                        )
                        self.result.type2.init_group()
                        self.result.type2.group.store(
                            'type2', self.cfg, self.data, self.compute_type2_negll,
                            params=[fit_type2_group.x[s] for s in range(self.data.nsubjects)],
                            params_se=[fit_type2_group.x_se[s] for s in range(self.data.nsubjects)],
                            params_cov=None if fit_type2_group.x_cov is None else [fit_type2_group.x_cov[s] for s in range(self.data.nsubjects)],
                            pop_mean_sd=fit_type2_group.x_re_pop_mean_sd,
                            execution_time=fit_type2_group.execution_time
                        )

            self.result.type2.store(cfg=self.cfg, data=self.data, fun=self.compute_type2_negll)
            if verbosity:
                self.result.type2.report_fit(self.cfg)

        self.result.store(store_type1_only=False)
        self.type2_is_fitted = True

        # if not silence_warnings:
        #     print_warnings(w)
        if verbosity:
            print('Type 2 level finished')

        return self

    def summary(self, generative=False, generative_nsamples=1000, squeeze=True):
        """
        Provides information about the model fit.

        Parameters
        ----------
        generative : bool
            If True, compare model predictions of confidence with empirical confidence by repeatedly sampling from
            the generative model.
        generative_nsamples : int
            Number of samples used for the generative model (higher = more accurate).
        squeeze : bool (default: True)
            If True, return flattened results in case of only a single participant.

        Returns
        ----------
        summary : dataclass
            Information about model fit.
        """

        if self.type2_is_fitted and generative:
            c_conf_generative = [np.empty((generative_nsamples, self.data.nsamples[s])) for s in range(self.data.nsubjects)]
            for s in range(self.data.nsubjects):
                c_conf_generative[s] = simulate(self.result.params[s], nsubjects=generative_nsamples, nsamples=self.data.nsamples[s],
                                                cfg=self.cfg, custom_stimuli=self.data.x_stim[s], verbosity=0).confidence
        else:
            c_conf_generative = None
        model_summary = self.result.summary(
            c_conf_empirical=self.data.c_conf, c_conf_generative=c_conf_generative, squeeze=squeeze
        )
        return model_summary


    def compute_type1_negll(self, params, sub_ind=0, save_type=None):
        """
        Likelihood function for the type 1 level

        Parameters:
        -----------
        params : array-like of shape (nparams)
            Parameter array of the type 1 level.
        sub_ind : int
            Subject index (only valid for 2d multi-subject datasets)
        save_type : None | str
            If 'subject' or 'group', store latent variables and parameters.

        Returns:
        --------
        negll: float
            Negative (summed) log likelihood.
        """

        # bl = self.cfg._paramset_type1.param_len_list
        # params_type1 = {p: params[int(np.sum(bl[:i]))] if n == 1 else [params[int(np.sum(bl[:i])) + j] for j in range(n)]
        #                 for i, (p, n) in enumerate(zip(self.cfg._paramset_type1.param_names, bl))}
        params_type1 = {p: params[ind] for p, ind in self.cfg._paramset_type1.param_ind.items()}

        type1_thresh = _check_param(params_type1['type1_thresh'] if self.cfg.param_type1_thresh.enable else self.cfg.param_type1_thresh.default)
        type1_bias = _check_param(params_type1['type1_bias'] if self.cfg.param_type1_bias.enable else self.cfg.param_type1_bias.default)

        if self.cfg.param_type1_nonlinear_gain.enable:
            x_stim_transform = compute_nonlinear_encoding(
                self.data.x_stim[sub_ind], params_type1['type1_nonlinear_gain'],
                params_type1['type1_nonlinear_scale'] if self.cfg.param_type1_nonlinear_scale.enable else
                (self.data.x_stim_max if self.cfg.param_type1_nonlinear_scale.default is None else self.cfg.param_type1_nonlinear_scale.default))
        else:
            x_stim_transform = self.data.x_stim[sub_ind]

        cond_neg, cond_pos = self.data.x_stim[sub_ind] < 0, self.data.x_stim[sub_ind] >= 0
        y_decval = np.full(self.data.x_stim[sub_ind].shape, np.nan)
        y_decval[cond_neg] = (np.abs(x_stim_transform[cond_neg]) > type1_thresh[0]) * x_stim_transform[cond_neg] + type1_bias[0]
        y_decval[cond_pos] = (np.abs(x_stim_transform[cond_pos]) > type1_thresh[1]) * x_stim_transform[cond_pos] + type1_bias[1]

        if self.cfg.param_type1_noise_heteroscedastic.enable or (self.cfg.param_type1_noise.enable == 2):
            type1_noise = compute_signal_dependent_type1_noise(
                x_stim=x_stim_transform,
                type1_noise_signal_dependency=self.cfg.param_type1_noise_heteroscedastic.model if self.cfg.param_type1_noise_heteroscedastic.enable else None,
                **params_type1
            )
        else:
            type1_noise = params_type1['type1_noise']

        if self.cfg.param_type1_noise.model == 'normal':
            posterior = ndtr(y_decval / type1_noise)
        elif self.cfg.param_type1_noise.model == 'logistic':
            posterior = logistic(y_decval, type1_noise)
        likelihood = (self.data.d_dec[sub_ind] == 1) * posterior + (self.data.d_dec[sub_ind] == 0) * (1 - posterior)
        negll = np.sum(-np.log(np.maximum(likelihood, self.cfg.min_type1_like)))

        # Add negative log likelihood of (fixed) Normal priors
        priors = [((params_type1[k] - p.prior[0])**2) / (2 * p.prior[1]**2) for k, p in
                  self.cfg._paramset_type1.parameters_flat.items() if isinstance(p.prior, tuple)]
        if len(priors) > 0:
            negll += np.sum(priors)

        if save_type is not None and (save_type != 'mock'):
            getattr(self.result.type1, save_type).params[sub_ind] = params_type1
            getattr(self.result.type1, save_type).loglik[sub_ind] = -negll
            if self.modeldata.type1_posterior is None:
                self.modeldata.type1_posterior = empty_list(self.data.nsubjects, self.data.nsamples)
                self.modeldata.type1_likelihood = empty_list(self.data.nsubjects, self.data.nsamples)
            self.modeldata.type1_posterior[sub_ind] = posterior
            self.modeldata.type1_likelihood[sub_ind] = likelihood
        return negll

    def compute_type2_negll(self, params, sub_ind=0, type2_noise_type=None, save_type=None):
        """
        Negative log likelihood minimization

        Parameters:
        -----------
        params : array-like of shape (nparams)
            Parameter array of the type 2 level.
        sub_ind : int
            Subject index (only valid for 2d multi-subject datasets)
        type2_noise_type : str
            Type 2 noise type: 'report', 'readout' or 'temperature'
        save_type : None | str
            If 'subject' or 'group', compute proper likelihoods and store latent variables and parameters.
            If 'mock', compute proper likelihoods but don't store latent variables and parameters.

        Returns:
        --------
        By default, the method returns the negative log likelihood. However, depending on the arguments various
        combinations of variables are returned (see Parameters).
        negll: float
            Negative (summed) log likelihood.
        """

        type2_noise_type = self.cfg.type2_noise_type if type2_noise_type is None else type2_noise_type

        # bl = self.cfg._paramset_type2.param_len_list
        # params_type2 = {p: params[int(np.sum(bl[:i]))] if n == 1 else [params[int(np.sum(bl[:i])) + j] for j in range(n)]
        #                 for i, (p, n) in enumerate(zip(self.cfg._paramset_type2.param_names, bl))}
        params_type2 = {p: params[ind] for p, ind in self.cfg._paramset_type2.param_ind.items()}

        if self.cfg.param_type2_criteria.enable or self.cfg.param_type2_criteria.preset is not None:
            likelihood_grid = self._compute_type2_likelihood_criteria(params_type2, sub_ind, type2_noise_type=type2_noise_type)
        else:
            likelihood_grid = self._compute_type2_likelihood_continuous(params_type2, sub_ind, type2_noise_type=type2_noise_type)

        # if save_type is not None and ('type2_criteria' in params_type2) and (np.any(params_type2['type2_criteria']) > 1.001):
        #     params_type2['type2_criteria'] = check_criteria(params_type2['type2_criteria'])

        if type2_noise_type == 'temperature':
            likelihood = likelihood_grid  # in case of the noisy-temp model there is no grid
            likelihood_grid = None
        else:
            if not self.cfg.type1_likel_incongr:
                likelihood_grid[self.modeldata.y_decval_grid_invalid[sub_ind]] = np.nan
            # compute log likelihood
            likelihood = np.nansum(self.modeldata.y_decval_pmf_grid[sub_ind] * likelihood_grid, axis=1)
            # This is equivalent:
            # type2_cum_likelihood2 = np.trapezoid(self.model.y_decval_grid_pdf * np.nan_to_num(likelihood_grid, 0),
            #                                      self.model.y_decval_grid)

        if self.cfg.min_type2_like_uni:
            # use an upper bound for the negative log likelihood based on a uniform 'guessing' model
            negll = min(self.modeldata.precomputed.uniform_type2_negll[sub_ind], -np.sum(np.log(np.maximum(likelihood, 1e-200))))
        else:
            negll = -np.sum(np.log(np.maximum(likelihood, self.cfg.min_type2_like)))

        if save_type is None:
            # We add likelihood punishments only during fitting, not for the final likelihood computation
            # Add likelihood punishment for invalid criteria
            if self.cfg.param_type2_criteria.enable and np.any((crit_abs := np.cumsum(params_type2['type2_criteria'])) > 1):
                negll *= crit_abs[crit_abs > 1].mean()

        # Add negative log likelihood of (fixed) Normal priors
        priors = [((params_type2[k] - p.prior[0])**2) / (2 * p.prior[1]**2) for k, p in
                  self.cfg._paramset_type2.parameters_flat.items() if isinstance(p.prior, tuple)]
        if len(priors) > 0:
            negll += np.sum(priors)

        if save_type is not None and (save_type != 'mock'):
            if type2_noise_type != 'temperature':
                if self.modeldata.c_conf_grid is None:
                    self.modeldata.c_conf_grid = empty_list(self.data.nsubjects, self.data.nsamples, self.cfg.type1_marg_steps)
                self.modeldata.c_conf_grid[sub_ind] = self._type1_evidence_to_confidence(self.modeldata.z1_type1_evidence_grid[sub_ind], params_type2, sub_ind)
                if not self.cfg.type1_likel_incongr:
                    self.modeldata.c_conf_grid[sub_ind][self.modeldata.y_decval_grid_invalid[sub_ind]] = np.nan
            if self.modeldata.c_conf is None:
                self.modeldata.c_conf = empty_list(self.data.nsubjects, self.data.nsamples)
            self.modeldata.c_conf[sub_ind] = self._type1_evidence_to_confidence(self.modeldata.z1_type1_evidence[sub_ind], params_type2, sub_ind)
            if self.modeldata.type2_likelihood is None:
                self.modeldata.type2_likelihood = empty_list(self.data.nsubjects, self.data.nsamples)
                self.modeldata.type2_likelihood_grid = empty_list(self.data.nsubjects, self.data.nsamples, self.cfg.type1_marg_steps)
            self.modeldata.type2_likelihood[sub_ind] = likelihood
            self.modeldata.type2_likelihood_grid[sub_ind] = likelihood_grid

            # Convert back to cumulative criteria!
            getattr(self.result.type2, save_type).params[sub_ind] = \
                {**params_type2, **dict(type2_criteria=np.minimum(1, np.cumsum(params_type2['type2_criteria'])))} \
                    if 'type2_criteria' in params_type2 else params_type2
            getattr(self.result.type2, save_type).loglik[sub_ind] = -negll

        return negll


    def _compute_type2_likelihood_continuous(self, params_type2, sub_ind, type2_noise_type):
        if self.cfg.type2_binsize_wrap:
            wrap_neg = (self.cfg.type2_binsize -
                        np.abs(np.minimum(1, self.data.c_conf_3d[sub_ind] + self.cfg.type2_binsize) - self.data.c_conf_3d[sub_ind]))
            wrap_pos = (self.cfg.type2_binsize -
                        np.abs(np.maximum(0, self.data.c_conf_3d[sub_ind] - self.cfg.type2_binsize) - self.data.c_conf_3d[sub_ind]))
            binsize_neg, binsize_pos = self.cfg.type2_binsize + wrap_neg, self.cfg.type2_binsize + wrap_pos
        else:
            binsize_neg, binsize_pos = self.cfg.type2_binsize, self.cfg.type2_binsize

        if type2_noise_type == 'temperature':
            if np.isclose(self.cfg.type2_binsize, 0):
                data = self.data.c_conf[sub_ind]
            else:
                data_lower = np.maximum(0, self.data.c_conf[sub_ind] - binsize_neg)
                data_upper = np.minimum(1, self.data.c_conf[sub_ind] + binsize_pos)
            dist = self.get_temperature_dist(params_type2)
        elif type2_noise_type == 'report':
            c_conf_grid = self._type1_evidence_to_confidence(self.modeldata.z1_type1_evidence_grid[sub_ind], params_type2, sub_ind)
            if np.isclose(self.cfg.type2_binsize, 0):
                data = self.data.c_conf_3d[sub_ind]
            else:
                data_lower = np.maximum(0, self.data.c_conf_3d[sub_ind] - binsize_neg)
                data_upper = np.minimum(1, self.data.c_conf_3d[sub_ind] + binsize_pos)
            dist = get_type2_dist(self.cfg.param_type2_noise.model, type2_center=c_conf_grid, type2_noise=params_type2['type2_noise'],
                                  type2_noise_type='report')
        elif type2_noise_type == 'readout':
            if np.isclose(self.cfg.type2_binsize, 0):
                data = self._confidence_to_type1_evidence(self.data.c_conf_3d[sub_ind], params_type2, sub_ind)
            else:
                data_lower = self._confidence_to_type1_evidence(
                    np.maximum(0, self.data.c_conf_3d[sub_ind] - binsize_neg), params_type2, sub_ind)
                data_upper = self._confidence_to_type1_evidence(
                    np.minimum(1, self.data.c_conf_3d[sub_ind] + binsize_pos), params_type2, sub_ind)
            dist = get_type2_dist(self.cfg.param_type2_noise.model, type2_center=self.modeldata.z1_type1_evidence_grid[sub_ind], type2_noise=params_type2['type2_noise'],
                                  type2_noise_type='readout')

        if np.isclose(self.cfg.type2_binsize, 0):
            type2_likelihood = dist.pdf(data)
        else:
            type2_likelihood = dist.cdf(data_upper) - dist.cdf(data_lower)

        return type2_likelihood

    def get_temperature_dist(self, params_type2, sub_ind=0, mask=None):

        def __init__(self_):

            self_.type2_dist = get_type2_dist(self.cfg.param_type2_noise.model, type2_center=self.result.type1.params[sub_ind]['type1_noise'], type2_noise=params_type2['type2_noise'],
                                              type2_noise_type='temperature')
            self_.params_type2 = params_type2
            # self_.type1_noise_range = self_.type2_dist.ppf(self.modeldata.precomputed_variables.quintiles_temperature)[:, None]
            # self_.type1_noise_range = np.maximum.accumulate(self_.type2_dist.ppf(self.modeldata.precomputed_variables.quintiles_temperature))[:, None]
            self_.type1_noise_range = self_.type2_dist.ppf(self.modeldata.precomputed.quintiles_temperature)
            if np.any(np.diff(self_.type1_noise_range) <= 0):
                raise ValueError('Numerical instability in the type 2 noise distribution. The lower bound'
                                 'of the type 2 noise parameter might be too small.')
            self_.type1_noise_range = self_.type1_noise_range[:, None]


        def conf_to_decval(self_, confidence):
            z1 = self._confidence_to_type1_evidence(confidence, self_.params_type2, type1_noise=self_.type1_noise_range,
                                                    tile_on_type1_uncertainty=False)
            y = z1 * (self.data.d_dec_sign[sub_ind] if mask is None else self.data.d_dec_sign[sub_ind][mask])
            return y

        def pdf(self_, data):
            y = self_.conf_to_decval(data)

            type2_evidence_bias = self_.params_type2['type2_evidence_bias'] if self.cfg.param_type2_evidence_bias.enable \
                else self.cfg.param_type2_evidence_bias.default
            if self.cfg.param.type2_confidence_bias.enable:
                pbias = self_.params_type2['type2_confidence_bias']
                c_conf = np.minimum(1-1e-8, self.data.c_conf[sub_ind])
                if self.cfg.param_type1_noise.model == 'normal':
                    # normal: c = 2*ɸ(z) - 1 -> z = ɸ^-1(0.5*(c+1)) -> Jacobian = dz/dc = sqrt(π/2)*exp((erf^-1(c))**2)
                    jac = pbias*(c_conf**(pbias-1))*np.sqrt(np.pi / 2.0) * np.exp(erfinv(c_conf**pbias)**2) / type2_evidence_bias
                elif self.cfg.param_type1_noise.model == 'logistic':
                    # logistic: c = tanh(z * (π/2√3)) -> z = (2√3/π) * atanh(c) -> Jacobian = dz/dc = (2√3/π) / (1 - c**2)
                    jac = ((2 * np.sqrt(3) * pbias * (c_conf**(pbias-1))) / np.pi) / (1 - c_conf**(2*pbias)) / type2_evidence_bias
            else:
                jac = self.modeldata.self.precomputed.jacobian_temperature[sub_ind] / type2_evidence_bias

            if self.cfg.param_type1_noise.model == 'normal':
                # Normal (Gaussian) pdf at y with mean mu and std sigma
                z = (y - self.modeldata.y_decval[sub_ind]) / self.result.type1.params[sub_ind]['type1_noise']
                pdf_y = np.exp(-0.5 * z * z) / (self.result.type1.params[sub_ind]['type1_noise'] * np.sqrt(2.0 * np.pi))
            elif self.cfg.param_type1_noise.model == 'logistic':
                # pdf_y = self.model.type1_dist.pdf(y)
                # slightly faster:
                ez = expit((y - self.modeldata.y_decval[sub_ind]) / (self.result.type1.params[sub_ind]['type1_noise'] * np.sqrt(3) / np.pi))
                pdf_y = (ez * (1 - ez)) / (self.result.type1.params[sub_ind]['type1_noise'] * np.sqrt(3) / np.pi)
            integrand = self_.type1_noise_range * self_.type2_dist.pdf(self_.type1_noise_range) * pdf_y
            pdf_ = jac * np.trapezoid(integrand, x=self_.type1_noise_range.squeeze(), axis=0)
            return pdf_

        def cdf(self_, data):
            y = self_.conf_to_decval(data)
            # cdf_y = self.model.type1_dist.cdf(y) - self.model.type1_dist.cdf(0) # slower!
            y_decval = self.modeldata.y_decval[sub_ind] if mask is None else self.modeldata.y_decval[sub_ind][mask]
            if self.cfg.param_type1_noise.model == 'normal':
                cdf_y = np.abs(ndtr(y_decval / self.result.type1.params[sub_ind]['type1_noise']) -
                               ndtr((y_decval - y) / self.result.type1.params[sub_ind]['type1_noise']))
            elif self.cfg.param_type1_noise.model == 'logistic':
                cdf_y = np.abs(logistic(y_decval, self.result.type1.params[sub_ind]['type1_noise']) -
                               logistic(y_decval - y, self.result.type1.params[sub_ind]['type1_noise']))
            # (999, 13)
            # neg = y < 0
            # if neg.sum():
            #     cdf_y[neg] = -cdf_y[neg]
            # Marginalize type 1 noise
            integrand = cdf_y * self_.type2_dist.pdf(self_.type1_noise_range)
            cdf_ = np.trapezoid(integrand, x=self_.type1_noise_range.squeeze(), axis=0)
            return cdf_

        Dist = type("Dist", (), {"__init__": __init__, "pdf": pdf, "cdf": cdf, "conf_to_decval": conf_to_decval})

        return Dist()


    def _compute_type2_likelihood_criteria(self, params_type2, sub_ind, type2_noise_type):

        type2_likelihood = np.empty_like(self.modeldata.y_decval[sub_ind], float) if type2_noise_type == 'temperature' \
                      else np.empty_like(self.modeldata.z1_type1_evidence_grid[sub_ind], float)

        if self.cfg.param_type2_criteria.enable:
            criteria_ = np.hstack((params_type2['type2_criteria'], 1))
        elif self.cfg.param_type2_criteria.preset is not None:
            # Convert to criteria gaps
            criteria_ = np.hstack((np.diff(np.hstack((0, self.cfg.param_type2_criteria.preset))), 1))
        else:
            raise ValueError('Type2 criteria are neither fitted nor predefined.')

        for i, crit in enumerate(criteria_):
            cnd = self.data.c_conf_discrete[sub_ind] == i
            if cnd.sum():
                if i == 0:
                    lower_bin_edge = 0
                    upper_bin_edge = crit
                else:
                    lower_bin_edge = np.sum(criteria_[:i])
                    upper_bin_edge = np.sum(criteria_[:i+1])
                if type2_noise_type == 'report':
                    c_conf_grid = self._type1_evidence_to_confidence(self.modeldata.z1_type1_evidence_grid[sub_ind], params_type2, sub_ind)
                    data_lower = lower_bin_edge
                    data_upper = min(1, upper_bin_edge)
                    type2_noise_dist = get_type2_dist(
                        self.cfg.param_type2_noise.model, type2_center=c_conf_grid[cnd], type2_noise=params_type2['type2_noise'],
                        type2_noise_type='report'
                    )
                elif type2_noise_type == 'readout':
                    data_lower = self._confidence_to_type1_evidence(lower_bin_edge, params_type2, sub_ind, mask=cnd)
                    data_upper = self._confidence_to_type1_evidence(upper_bin_edge, params_type2, sub_ind, mask=cnd)
                    type2_noise_dist = get_type2_dist(
                        self.cfg.param_type2_noise.model, type2_center=self.modeldata.z1_type1_evidence_grid[sub_ind][cnd], type2_noise=params_type2['type2_noise'],
                        type2_noise_type='readout'
                    )
                elif type2_noise_type == 'temperature':
                    data_lower = lower_bin_edge
                    data_upper = min(1, upper_bin_edge)
                    type2_noise_dist = self.get_temperature_dist(params_type2, sub_ind, mask=cnd)

                type2_likelihood[cnd] = type2_noise_dist.cdf(data_upper) - type2_noise_dist.cdf(data_lower)  # noqa

        return type2_likelihood


    def _type1_evidence_to_confidence(self, z1_type1_evidence, params_type2, sub_ind=0):
        """
        Helper function to convert type 1 evidence to confidence
        """
        return type1_evidence_to_confidence(
            z1_type1_evidence=z1_type1_evidence, type1_dist=self.cfg.param_type1_noise.model,
            x_stim=self.data.x_stim[sub_ind],
            type1_noise_signal_dependency=self.cfg.param_type1_noise_heteroscedastic.model if self.cfg.param_type1_noise_heteroscedastic.enable else None,
            **self.result.type1.params[sub_ind], **params_type2
        )

    def _type1_noise_to_confidence(self, type1_noise, params_type2, sub_ind=0):
        """
        Helper function to convert type 1 noise to confidence
        """
        return type1_noise_to_confidence(
            y_decval=self.modeldata.y_decval_grid[sub_ind], type1_dist=self.cfg.param_type1_noise.model,
            x_stim=self.data.x_stim[sub_ind],
            type1_noise_signal_dependency=self.cfg.param_type1_noise_heteroscedastic.model if self.cfg.param_type1_noise_heteroscedastic.enable else None,
            **{**self.result.type1.params[sub_ind], **dict(type1_noise=type1_noise)}, **params_type2
        )

    def _confidence_to_type1_evidence(self, c_conf, params_type2, sub_ind=0, type1_noise=None,
                                      tile_on_type1_uncertainty=True, mask=None):
        """
        Helper function to convert confidence to type 1 evidence
        """
        return confidence_to_type1_evidence(
            c_conf=c_conf, type1_dist=self.cfg.param_type1_noise.model,
            x_stim=self.data.x_stim[sub_ind] if mask is None else self.data.x_stim[sub_ind][mask],
            y_decval=self.modeldata.y_decval_grid[sub_ind] if mask is None else self.modeldata.y_decval_grid[sub_ind][mask],
            type1_noise_signal_dependency=self.cfg.param_type1_noise_heteroscedastic.model if self.cfg.param_type1_noise_heteroscedastic.enable else None,
            tile_on_type1_uncertainty=tile_on_type1_uncertainty,
            **(self.result.type1.params[sub_ind] if type1_noise is None else
               {**self.result.type1.params[sub_ind], **dict(type1_noise=type1_noise)}),
            **params_type2
        )

    def _confidence_to_type1_noise(self, c_conf, params_type2, sub_ind=0, mask=None):
        """
        Helper function to convert confidence to type 1 noise
        """
        return confidence_to_type1_noise(
            c_conf=c_conf, type1_dist=self.cfg.param_type1_noise.model,
            x_stim=self.data.x_stim[sub_ind] if mask is None else self.data.x_stim[sub_ind][mask],
            y_decval=self.modeldata.y_decval_grid[sub_ind] if mask is None else self.modeldata.y_decval_grid[sub_ind][mask],
            type1_noise_signal_dependency=self.cfg.param_type1_noise_heteroscedastic.model if self.cfg.param_type1_noise_heteroscedastic.enable else None,
            **self.result.type1.params[sub_ind], **params_type2
        )

    def _compute_decision_values(self):
        """
        Compute type 1 decision values
        """

        range_ = np.linspace(0, self.cfg.type1_marg_z, int((self.cfg.type1_marg_steps + 1) / 2))[1:]
        yrange = np.hstack((-range_[::-1], 0, range_))
        self.modeldata.y_decval_grid = [np.empty((self.data.nsamples[s], yrange.shape[0])) for s in range(self.data.nsubjects)]
        self.modeldata.y_decval = [np.empty(self.data.nsamples[s]) for s in range(self.data.nsubjects)]
        if self.cfg.type2_noise_type != 'temperature':
            self.modeldata.y_decval_pmf_grid = [np.empty(self.modeldata.y_decval_grid[s].shape) for s in range(self.data.nsubjects)]
            if not self.cfg.type1_likel_incongr:
                self.modeldata.y_decval_grid_invalid = [np.empty(self.modeldata.y_decval_grid[s].shape, dtype=bool) for s in range(self.data.nsubjects)]
        if self.cfg.min_type2_like_uni:
            self.modeldata.precomputed.uniform_type2_negll = np.empty(self.data.nsubjects)
        for s in range(self.data.nsubjects):

            type1_noise_trialwise = compute_signal_dependent_type1_noise(
                x_stim=self.data.x_stim_3d[s],
                type1_noise_signal_dependency=self.cfg.param_type1_noise_heteroscedastic.model if self.cfg.param_type1_noise_heteroscedastic.enable else None,
                **self.result.type1.params[s]
            )
            type1_thresh = _check_param(self.result.type1.params[s]['type1_thresh'] if self.cfg.param_type1_thresh.enable else 0)
            type1_bias = _check_param(self.result.type1.params[s]['type1_bias'] if self.cfg.param_type1_bias.enable else 0)

            cond_neg, cond_pos = (self.data.x_stim_3d[s] < 0).squeeze(), (self.data.x_stim_3d[s] >= 0).squeeze()
            y_decval = np.empty(self.data.x_stim_3d[s].shape)
            y_decval[cond_neg] = (np.abs(self.data.x_stim_3d[s][cond_neg]) >= type1_thresh[0]) * \
                self.data.x_stim_3d[s][cond_neg] + type1_bias[0]
            y_decval[cond_pos] = (np.abs(self.data.x_stim_3d[s][cond_pos]) >= type1_thresh[1]) * \
                self.data.x_stim_3d[s][cond_pos] + type1_bias[1]
            self.modeldata.y_decval[s] = y_decval.flatten()

            self.modeldata.y_decval_grid[s][cond_neg] = y_decval[cond_neg] + yrange * np.mean(type1_noise_trialwise[cond_neg])
            self.modeldata.y_decval_grid[s][cond_pos] = y_decval[cond_pos] + yrange * np.mean(type1_noise_trialwise[cond_pos])

            if self.cfg.type2_noise_type != 'temperature':

                margin_neg = type1_noise_trialwise[cond_neg] * self.cfg.type1_marg_z / self.cfg.type1_marg_steps
                margin_pos = type1_noise_trialwise[cond_pos] * self.cfg.type1_marg_z / self.cfg.type1_marg_steps

                if self.cfg.param_type1_noise.model == 'normal':
                    self.modeldata.y_decval_pmf_grid[s][cond_neg] = (
                        ndtr((self.modeldata.y_decval_grid[s][cond_neg] + margin_neg - y_decval[cond_neg]) / type1_noise_trialwise[cond_neg]) -
                        ndtr((self.modeldata.y_decval_grid[s][cond_neg] - margin_neg - y_decval[cond_neg]) / type1_noise_trialwise[cond_neg])
                    )
                    self.modeldata.y_decval_pmf_grid[s][cond_pos] = (
                        ndtr((self.modeldata.y_decval_grid[s][cond_pos] + margin_pos - y_decval[cond_pos]) / type1_noise_trialwise[cond_pos]) -
                        ndtr((self.modeldata.y_decval_grid[s][cond_pos] - margin_pos - y_decval[cond_pos]) / type1_noise_trialwise[cond_pos])
                    )
                elif self.cfg.param_type1_noise.model == 'logistic':
                    self.modeldata.y_decval_pmf_grid[s][cond_neg] = (
                        logistic(self.modeldata.y_decval_grid[s][cond_neg] + margin_neg - y_decval[cond_neg], type1_noise_trialwise[cond_neg]) -
                        logistic(self.modeldata.y_decval_grid[s][cond_neg] - margin_neg - y_decval[cond_neg], type1_noise_trialwise[cond_neg])
                    )
                    self.modeldata.y_decval_pmf_grid[s][cond_pos] = (
                        logistic(self.modeldata.y_decval_grid[s][cond_pos] + margin_pos - y_decval[cond_pos], type1_noise_trialwise[cond_pos]) -
                        logistic(self.modeldata.y_decval_grid[s][cond_pos] - margin_pos - y_decval[cond_pos], type1_noise_trialwise[cond_pos])
                    )
                    # cdf_neg = logistic_dist(loc=y_decval[cond_neg], scale=type1_noise_trialwise[cond_neg] * np.sqrt(3) / np.pi)
                    # cdf_pos = logistic_dist(loc=y_decval[cond_pos], scale=type1_noise_trialwise[cond_pos] * np.sqrt(3) / np.pi)
                    # self.modeldata.y_decval_pmf_grid[s][cond_neg] = (cdf_neg.cdf(self.modeldata.y_decval_grid[s][cond_neg] + margin_neg) -
                    #                                                  cdf_neg.cdf(self.modeldata.y_decval_grid[s][cond_neg] - margin_neg))
                    # self.modeldata.y_decval_pmf_grid[s][cond_pos] = (cdf_pos.cdf(self.modeldata.y_decval_grid[s][cond_pos] + margin_pos) -
                    #                                                  cdf_pos.cdf(self.modeldata.y_decval_grid[s][cond_pos] - margin_pos))

                # normalize PMF
                self.modeldata.y_decval_pmf_grid[s] = self.modeldata.y_decval_pmf_grid[s] / self.modeldata.y_decval_pmf_grid[s].sum(axis=-1).reshape(-1, 1)
                # invalidate invalid decision values
                if not self.cfg.type1_likel_incongr:
                    self.modeldata.y_decval_grid_invalid[s] = np.sign(self.modeldata.y_decval_grid[s]) != \
                                                              np.sign(self.data.d_dec_3d[s] - 0.5)
                    self.modeldata.y_decval_pmf_grid[s][self.modeldata.y_decval_grid_invalid[s]] = np.nan

                if self.cfg.min_type2_like_uni:
                    # self.cfg.type2_binsize*2 is the probability for a given confidence rating assuming a uniform
                    # distribution for confidence. This 'confidence guessing model' serves as a upper bound for the
                    # negative log likelihood.
                    min_type2_likelihood_grid = self.cfg.type2_binsize * 2 * np.ones(self.modeldata.y_decval_pmf_grid[s].shape)
                    min_type2_likelihood = np.nansum(min_type2_likelihood_grid * self.modeldata.y_decval_pmf_grid[s], axis=1)
                    self.modeldata.precomputed.uniform_type2_negll[s] = -np.log(min_type2_likelihood).sum()

        if self.cfg.type2_noise_type != 'temperature':
            self.modeldata.z1_type1_evidence_grid = [np.abs(self.modeldata.y_decval_grid[s]) for s in range(self.data.nsubjects)]

        self.modeldata.z1_type1_evidence = [np.abs(self.modeldata.y_decval[s]) for s in range(self.data.nsubjects)]


    def _check_fit(self):
        if not self.type1_is_fitted and not self.type2_is_fitted:
            raise RuntimeError('Please fit the model before plotting.')
        elif not self.cfg.skip_type2 and (self.type1_is_fitted and not self.type2_is_fitted):
            raise RuntimeError('Only the type 1 level was fitted. Please also fit the type 2 level to plot'
                               'a link function.')

    def plot_psychometric(
        self,
        sub_ind: int = 0,
        model_prediction: bool = True,
        **kwargs
    ):
        """
        Invoke `remeta.plotting.plot_psychomtric` on the `ReMeta` instance.

        Usage:
            ```
            rem = ReMeta()
            rem.fit(stimuli, choices, confidence)
            rem.plot_psychometric()
            ```

        Args:
            sub_ind: subject index (only applicable if group data were fitted)
            model_prediction: if `True`, plot model predictions
            **kwargs: Pass parameters to `remeta.plotting.plot_psychomtric`
        """

        # self._check_fit()
        plot_psychometric(
            self.data.x_stim[sub_ind], self.data.d_dec[sub_ind],
            params=self.result.params[sub_ind] if model_prediction else None,
            cfg=self.cfg, **kwargs
        )

    def plot_stimulus_versus_confidence(
        self,
        sub_ind: int = 0,
        model_prediction: bool = True,
        **kwargs
    ):
        """
        Invoke `remeta.plotting.plot_stimulus_versus_confidence` on the `ReMeta` instance.

        Usage:
            ```
            rem = ReMeta()
            rem.fit(stimuli, choices, confidence)
            rem.plot_stimulus_versus_confidence()
            ```

        Args:
            sub_ind: subject index (only applicable if group data were fitted)
            model_prediction: if `True`, plot model predictions
            **kwargs: Pass parameters to `remeta.plotting.plot_stimulus_versus_confidence`
        """
        # self._check_fit()
        plot_stimulus_versus_confidence(
            self.data.x_stim[sub_ind], self.data.c_conf[sub_ind],self.data.d_dec[sub_ind],
            params=self.result.params[sub_ind] if model_prediction else None,
            model_prediction=model_prediction,
            cfg=self.cfg, **kwargs
        )

    def plot_confidence_histogram(
        self,
        sub_ind: int = 0,
        model_prediction: bool = True,
        **kwargs
    ):
        """
        Invoke `remeta.plotting.plot_confidence_histogram` on the `ReMeta` instance.

        Usage:
            ```
            rem = ReMeta()
            rem.fit(stimuli, choices, confidence)
            rem.plot_confidence_histogram()
            ```

        Args:
            sub_ind: subject index (only applicable if group data were fitted)
            model_prediction: if `True`, plot model predictions
            **kwargs: Pass parameters to `remeta.plotting.plot_confidence_histogram`
        """
        # self._check_fit()
        plot_confidence_histogram(
            self.data.c_conf[sub_ind], self.data.x_stim[sub_ind], self.data.d_dec[sub_ind],
            params=self.result.params[sub_ind] if model_prediction else None,
            cfg=self.cfg, **kwargs
        )


def load_dataset(name, verbosity=1, return_data_only=False):
    import gzip
    path = os.path.join(pathlib.Path(__file__).parent.resolve(), 'demo_data', f'example_data_{name}.pkl.gz')
    if os.path.exists(path):
        with gzip.open(path, 'rb') as f:
            dataset = pickle.load(f)
    else:
        raise FileNotFoundError(f'[Dataset does not exist!] No such file: {path}')

    if verbosity:
        print_dataset_characteristics(dataset)

    if return_data_only:
        return (dataset.x_stim, dataset.d_dec, dataset.params)
    else:
        return dataset


import numpy as np
from scipy.stats import norm
from scipy.optimize import minimize
import warnings

# Python translation from Brian Maniscalco's code at http://www.columbia.edu/~bsm2105/type2sdt/
# Matthias Guggenmos, 2021


def fit_meta_d_MLE(nR_S1, nR_S2, s=1, fncdf=None, fninv=None):
    # fit = fit_meta_d_MLE(nR_S1, nR_S2, s, fncdf, fninv)
    #
    # Given data from an experiment where an observer discriminates between two
    # stimulus alternatives on every trial and provides confidence ratings,
    # provides a type 2 SDT analysis of the data.
    #
    # INPUTS

    # * nR_S1, nR_S2
    # these are vectors containing the total number of responses in
    # each response modality, conditional on presentation of S1 and S2.
    #
    # e.g. if nR_S1 = [100 50 20 10 5 1], then when stimulus S1 was
    # presented, the subject had the following response counts:
    # responded S1, rating=3 : 100 times
    # responded S1, rating=2 : 50 times
    # responded S1, rating=1 : 20 times
    # responded S2, rating=1 : 10 times
    # responded S2, rating=2 : 5 times
    # responded S2, rating=3 : 1 time
    #
    # * s
    # this is the ratio of standard deviations for type 1 distributions, i.e.
    #
    # s = sd(S1) / sd(S2)
    #
    # if not specified, s is set to a default value of 1.
    # the function SDT_MLE_fit available at
    # http://www.columbia.edu/~bsm2105/type2sdt/
    # can be used to get an MLE estimate of s using Matlab's optimization
    # toolbox
    #
    # * fncdf
    # a function handle for the CDF of the type 1 distribution.
    # if not specified, fncdf defaults to @normcdf (i.e. CDF for normal
    # distribution)
    #
    # * fninv
    # a function handle for the inverse CDF of the type 1 distribution.
    # if not specified, fninv defaults to @norminv
    #
    # OUTPUT
    #
    # Output is packaged in the struct "fit."
    # In the following, let S1 and S2 represent the distributions of evidence
    # generated by stimulus classes S1 and S2.
    # Then the fields of "fit" are as follows:
    #
    # fit.da        = mean(S2) - mean(S1), in room-mean-square(sd(S1),sd(S2)) units
    # fit.s         = sd(S1) / sd(S2)
    # fit.meta_da   = meta-d' in RMS units
    # fit.M_diff    = meta_da - da
    # fit.M_ratio   = meta_da / da
    # fit.t1ca      = type 1 criterion for meta-d' fit, RMS units
    # fit.t2ca      = type 2 criteria for meta-d' fit, RMS units
    #
    # fit.S1units   = contains same parameters in sd(S1) units.
    #                 these may be of use since the data-fitting is conducted
    #                 using parameters specified in sd(S1) units.
    #
    # fit.logL          = log likelihood of the data fit
    #
    # fit.obs_HR2_rS1  = actual type 2 hit rates for S1 responses
    # fit.est_HR2_rS1  = estimated type 2 hit rates for S1 responses
    # fit.obs_FAR2_rS1 = actual type 2 false alarm rates for S1 responses
    # fit.est_FAR2_rS1 = estimated type 2 false alarm rates for S1 responses
    #
    # fit.obs_HR2_rS2  = actual type 2 hit rates for S2 responses
    # fit.est_HR2_rS2  = estimated type 2 hit rates for S2 responses
    # fit.obs_FAR2_rS2 = actual type 2 false alarm rates for S2 responses
    # fit.est_FAR2_rS2 = estimated type 2 false alarm rates for S2 responses
    #
    # If there are N ratings, then there will be N-1 type 2 hit rates and false
    # alarm rates.

    # 9/7/10 - bm - wrote it

    ## parse inputs

    # nR_S1 = np.array([171275, 174686, 110922, 43117])
    # nR_S2 = np.array([43117, 111053, 174370, 171460])
    # nR_S1 = np.array([17664, 65846, 87765, 88184, 86502, 64187, 46735, 29635, 12072, 1410])
    # nR_S2 = np.array([1421, 11891, 29805, 46620, 64433, 86248, 88122, 88348, 65901, 17211])

    if not (np.mod(len(nR_S1), 2) == 0):
        raise Exception('input arrays must have an even number of elements')
    if len(nR_S1) != len(nR_S2):
        raise Exception('input arrays must have the same number of elements')

    if fncdf is None:
        fncdf = norm.cdf

    if fninv is None:
        fninv = norm.ppf

    nRatings = int(len(nR_S1) / 2)
    nCriteria = 2 * nRatings - 1

    ## set up constraints for MLE estimation

    # parameters
    # meta-d' - 1
    # t2c     - nCriteria-1

    # constrain type 2 criteria values,
    # such that t2c(i) is always <= t2c(i+1)
    # want t2c(i)   <= t2c(i+1)
    # -->  t2c(i+1) >= c(i) + 1e-5 (i.e. very small deviation from equality)
    # -->  t2c(i) - t2c(i+1) <= -1e-5
    A = np.zeros((nCriteria - 2, nCriteria))
    b = np.full(nCriteria - 2, np.nan)
    for i in range(nCriteria - 2):
        A[i, (i + 1):(i + 3)] = [1, -1]
        b[i] = -1e-5

    # lower bounds on parameters
    LB = np.hstack((-10, -20 * np.ones(int((nCriteria - 1) / 2)), np.zeros((int((nCriteria - 1) / 2)))))

    # upper bounds on parameters
    UB = np.hstack((10, np.zeros((int((nCriteria - 1) / 2))), 20 * np.ones(int((nCriteria - 1) / 2))))

    ## select constant criterion type

    constant_criterion = 'meta_d1 * (t1c1 / d1)'  # relative criterion

    ## set up initial guess at parameter values

    ratingHR = np.full(nRatings * 2 - 1, np.nan)
    ratingFAR = np.full(nRatings * 2 - 1, np.nan)
    for c in range(1, nRatings * 2):
        ratingHR[c - 1] = np.sum(nR_S2[c:]) / np.sum(nR_S2)
        ratingFAR[c - 1] = np.sum(nR_S1[c:]) / np.sum(nR_S1)

    t1_index = nRatings - 1
    t2_index = np.setdiff1d(range(2 * nRatings - 1), t1_index)

    d1 = (1 / s) * fninv(ratingHR[t1_index]) - fninv(ratingFAR[t1_index])
    if np.abs(d1) < 1e-4:
        if d1 == 0:
            d1 = np.random.choice([-1, 1]) * 1e-4
        else:
            d1 = np.sign(d1) * 1e-4
        # d1 = 1e-4
        small_d1 = 1
    else:
        small_d1 = 0
    meta_d1 = d1

    c1 = (-1 / (1 + s)) * (fninv(ratingHR) + fninv(ratingFAR))
    t1c1 = c1[t1_index]
    t2c1 = c1[t2_index]

    guess = np.hstack((meta_d1, t2c1 - eval(constant_criterion)))

    ## find the best fit for type 2 hits and FAs

    bounds = [(lb, ub) for lb, ub in zip(LB, UB)]
    constraint = {'type': 'ineq', 'fun': lambda x: b - np.dot(A, x).squeeze()}
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=(UserWarning, RuntimeWarning))
        # res = minimize(fit_meta_d_logL, guess, args=(nR_S1, nR_S2, nRatings, d1, t1c1, s, constant_criterion, fncdf, fninv), method='SLSQP', bounds=bounds, constraints=constraint, options=dict(maxiter=1e4, disp=False))  # trust-constr, L-BFGS-B, TNC, SLSQP
        args = [nR_S1, nR_S2, nRatings, d1, t1c1, s, constant_criterion, fncdf, fninv]
        res = minimize(fit_meta_d_logL, guess, args=tuple(args), method='SLSQP', bounds=bounds, constraints=constraint, options=dict(disp=False))  # trust-constr, L-BFGS-B, TNC, SLSQP

    meta_d1 = res.x[0]
    t2c1 = res.x[1:] + eval(constant_criterion)
    logL = -res.fun



    ## data is fit, now to package it...

    ## find observed t2FAR and t2HR

    # I_nR and C_nR are rating trial counts for incorrect and correct trials
    # element i corresponds to # (in)correct w/ rating i
    I_nR_rS2 = nR_S1[nRatings:]
    I_nR_rS1 = nR_S2[nRatings - 1::-1]

    C_nR_rS2 = nR_S2[nRatings:]
    C_nR_rS1 = nR_S1[nRatings - 1::-1]

    obs_FAR2_rS2, obs_HR2_rS2 = np.full(nRatings - 1, np.nan), np.full(nRatings - 1, np.nan)
    obs_FAR2_rS1, obs_HR2_rS1 = np.full(nRatings - 1, np.nan), np.full(nRatings - 1, np.nan)
    for i in range(1, nRatings):
        obs_FAR2_rS2[i - 1] = np.sum(I_nR_rS2[i:]) / np.sum(I_nR_rS2)
        obs_HR2_rS2[i - 1] = np.sum(C_nR_rS2[i:]) / np.sum(C_nR_rS2)

        obs_FAR2_rS1[i - 1] = np.sum(I_nR_rS1[i:]) / np.sum(I_nR_rS1)
        obs_HR2_rS1[i - 1] = np.sum(C_nR_rS1[i:]) / np.sum(C_nR_rS1)

    ## find estimated t2FAR and t2HR

    S1mu = -meta_d1 / 2
    S1sd = 1
    S2mu = meta_d1 / 2
    S2sd = S1sd / s

    mt1c1 = eval(constant_criterion)

    C_area_rS2 = 1 - fncdf(mt1c1, S2mu, S2sd)
    I_area_rS2 = 1 - fncdf(mt1c1, S1mu, S1sd)

    C_area_rS1 = fncdf(mt1c1, S1mu, S1sd)
    I_area_rS1 = fncdf(mt1c1, S2mu, S2sd)

    est_FAR2_rS2, est_HR2_rS2 = np.full(nRatings - 1, np.nan), np.full(nRatings - 1, np.nan)
    est_FAR2_rS1, est_HR2_rS1 = np.full(nRatings - 1, np.nan), np.full(nRatings - 1, np.nan)
    for i in range(0, nRatings - 1):
        t2c1_lower = t2c1[nRatings - i - 2]
        t2c1_upper = t2c1[nRatings - 1 + i]

        I_FAR_area_rS2 = 1 - fncdf(t2c1_upper, S1mu, S1sd)
        C_HR_area_rS2 = 1 - fncdf(t2c1_upper, S2mu, S2sd)

        I_FAR_area_rS1 = fncdf(t2c1_lower, S2mu, S2sd)
        C_HR_area_rS1 = fncdf(t2c1_lower, S1mu, S1sd)

        est_FAR2_rS2[i] = I_FAR_area_rS2 / I_area_rS2
        est_HR2_rS2[i] = C_HR_area_rS2 / C_area_rS2

        est_FAR2_rS1[i] = I_FAR_area_rS1 / I_area_rS1
        est_HR2_rS1[i] = C_HR_area_rS1 / C_area_rS1

    ## package output

    class Fit:
        def __init__(self):
            self.da = np.sqrt(2 / (1 + s ** 2)) * s * d1
            self.d1 = d1
            if np.abs(self.da) < 1e-4:
                if self.da == 0:
                    self.da = np.random.choice([-1, 1]) * 1e-4
                else:
                    self.da = np.sign(self.da) * 1e-4
                self.small_da = 1
            else:
                self.small_da = 0
            self.s = s
            self.meta_da = np.sqrt(2 / (1 + s ** 2)) * s * meta_d1
            if np.abs(self.meta_da) < 1e-4:
                self.meta_da = np.sign(self.meta_da) * 1e-4
                self.small_meta_da = 1
            else:
                self.small_meta_da = 0
            self.small_d1 = small_d1
            self.M_diff = self.meta_da - self.da
            self.M_ratio = self.meta_da / self.da
            self.meta_ca = (np.sqrt(2) * s / np.sqrt(1 + s ** 2)) * t1c1
            t2ca = (np.sqrt(2) * s / np.sqrt(1 + s ** 2)) * t2c1
            self.t2ca_rS1 = t2ca[:nRatings - 1]
            self.t2ca_rS2 = t2ca[nRatings - 1:]
            self.S1units = dict(d1=d1, meta_d1=meta_d1, s=1, meta_c1=t1c1, t2c1_rS1=t2c1[:nRatings - 1], t2c1_rS2=t2c1[nRatings - 1:])
            self.logL = logL
            self.est_HR2_rS1 = est_HR2_rS1
            self.obs_HR2_rS1 = obs_HR2_rS1
            self.est_FAR2_rS1 = est_FAR2_rS1
            self.obs_FAR2_rS1 = obs_FAR2_rS1
            self.est_HR2_rS2 = est_HR2_rS2
            self.obs_HR2_rS2 = obs_HR2_rS2
            self.est_FAR2_rS2 = est_FAR2_rS2
            self.obs_FAR2_rS2 = obs_FAR2_rS2
            self.fit_nfev = res.nfev
            self.fit_nit = res.nit
            self.fit_njev = res.njev
            self.fit_success = int(res.success)
            self.fit_fun = res.fun

    return Fit()


## function to find the likelihood of parameter values, given observed data
def fit_meta_d_logL(parameters, nR_S1, nR_S2, nRatings, d1, t1c1, s, constant_criterion, fncdf, fninv):
    # set up parameters
    meta_d1 = parameters[0]
    t2c1 = np.array(parameters[1:])

    # loads:
    # nR_S1 nR_S2 nRatings d1 t1c1 s constant_criterion fncdf fninv
    # nR_S1, nR_S2, nRatings, d1, t1c1, s, constant_criterion, fncdf, fninv = pickle.load(open('fit_meta_d_MLE.pkl', 'rb'))

    # define mean and SD of S1 and S2 distributions
    S1mu = -meta_d1 / 2
    S1sd = 1
    S2mu = meta_d1 / 2
    S2sd = S1sd / s

    # adjust so that the type 1 criterion is set at 0
    # (this is just to work with optimization toolbox constraints...
    #  to simplify defining the upper and lower bounds of type 2 criteria)
    S1mu = S1mu - eval(constant_criterion)
    S2mu = S2mu - eval(constant_criterion)

    t1c1 = 0

    ### set up MLE analysis

    # get type 2 response counts
    nC_rS1, nC_rS2 = np.full(nRatings, np.nan), np.full(nRatings, np.nan)
    nI_rS1, nI_rS2 = np.full(nRatings, np.nan), np.full(nRatings, np.nan)
    for i in range(nRatings):
        # S1 responses
        nC_rS1[i] = nR_S1[i]
        nI_rS1[i] = nR_S2[i]

        # S2 responses
        nC_rS2[i] = nR_S2[nRatings + i]
        nI_rS2[i] = nR_S1[nRatings + i]

    # get type 2 probabilities
    C_area_rS1 = fncdf(t1c1, S1mu, S1sd)
    I_area_rS1 = fncdf(t1c1, S2mu, S2sd)

    C_area_rS2 = 1 - fncdf(t1c1, S2mu, S2sd)
    I_area_rS2 = 1 - fncdf(t1c1, S1mu, S1sd)

    t2c1x = np.hstack((-np.inf, t2c1[0:nRatings - 1], t1c1, t2c1[nRatings - 1:], np.inf));

    prC_rS1, prC_rS2 = np.full(nRatings, np.nan), np.full(nRatings, np.nan)
    prI_rS1, prI_rS2 = np.full(nRatings, np.nan), np.full(nRatings, np.nan)
    for i in range(nRatings):
        prC_rS1[i] = (fncdf(t2c1x[i + 1], S1mu, S1sd) - fncdf(t2c1x[i], S1mu, S1sd)) / C_area_rS1
        prI_rS1[i] = (fncdf(t2c1x[i + 1], S2mu, S2sd) - fncdf(t2c1x[i], S2mu, S2sd)) / I_area_rS1

        prC_rS2[i] = ((1 - fncdf(t2c1x[nRatings + i], S2mu, S2sd)) - (1 - fncdf(t2c1x[nRatings + i + 1], S2mu, S2sd))) / C_area_rS2
        prI_rS2[i] = ((1 - fncdf(t2c1x[nRatings + i], S1mu, S1sd)) - (1 - fncdf(t2c1x[nRatings + i + 1], S1mu, S1sd))) / I_area_rS2

    # calculate logL
    logL = 0
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=RuntimeWarning)
        for i in range(nRatings):
            logL = logL + nC_rS1[i] * np.log(prC_rS1[i]) + nI_rS1[i] * np.log(prI_rS1[i]) + \
                   nC_rS2[i] * np.log(prC_rS2[i]) + nI_rS2[i] * np.log(prI_rS2[i])

    if np.isnan(logL):
        logL = -np.inf

    logL = -logL

    return logL


def type2_SDT_MLE(stimID, response, rating, nRatings, cellpadding=None, equalVariance=1):
    # out = type2_SDT(input)
    #
    # Given data from an experiment where an observer discriminates between two
    # stimulus alternatives on every trial and provides confidence ratings,
    # provides a type 2 SDT analysis of the data.
    #
    # The function estimates the parameters of the unequal variance SDT model,
    # and uses those estimates to find a maximum likelihood estimate of
    # meta-da.
    #
    # INPUTS
    #
    # format of the input may be either:
    #
    # 1) stimID, response, rating, nRatings, (cellpadding), (equalVariance)
    #    where each of the first 3 inputs is a 1xN vector describing the outcome
    #    of N trials. Contents of input should be as follows.
    #
    #    stimID   : 0=S1 stimulus presented, 1=S2 stimulus presented
    #    response : 0=subject responded S1, 1=subject responded S2
    #    rating   : values ranges from 1 to m where 1 is the lowest rating
    #               and m is the highest.
    #
    #               All trials where any of these prescribed ranges of values
    #               are violated are omitted from analysis.
    #
    #    nRatings : the number of ratings available to the subject (e.g. for a
    #               confidence scale of 1-4, nRatings=4).
    #    cellpadding : if any data cells (e.g. high confidence "S2" responses)
    #               are empty, then the value of cellpadding will be added
    #               to every data cell. If not specified, default = 1/(2*nRatings)
    #    equalVariance : if 1, force analysis to use the equal variance SDT
    #               model. If 0, use an estimate of s = sd(S1) / sd(S2) where
    #               s is the slope of the zROC data (estimated using MLE).
    #               If not specified, default = 0.
    #
    # 2) nR_S1, nR_S2, (cellpadding), (equalVariance)
    #    where these are vectors containing the total number of responses in
    #    each response modality, conditional on presentation of S1 and S2.
    #    size of each array is 2*nRatings, where each element corresponds to a
    #    count of responses in each response modality. Response categories are
    #    ordered as follows:
    #    highest conf "S1" ... lowest conf "S1", lowest conf "S2", ... highest conf "S2"
    #
    #    e.g. if nR_S1 = [100 50 20 10 5 1], then when stimulus S1 was
    #    presented, the subject had the following response counts:
    #    responded S1, rating=3 : 100 times
    #    responded S1, rating=2 : 50 times
    #    responded S1, rating=1 : 20 times
    #    responded S2, rating=1 : 10 times
    #    responded S2, rating=2 : 5 times
    #    responded S2, rating=3 : 1 time
    #
    #    cellpadding and equalVariance are defined as above.
    #
    #
    #
    #
    # OUTPUTS
    #
    # out.d_a       : d_a for input data. If s=1, d_a = d'
    # out.meta_d_a  : meta_d_a for input data
    # out.M_ratio   : meta_d_a / d_a; measure of metacognitive efficiency
    # out.M_diff    : meta_d_a - d_a; measure of metacognitive efficiency
    # out.c_a       : criterion c_a for input data. If s=1, c_a = c.
    # out.cprime    : relative criterion used for type 2 estimates. c' = c_a / d_a
    # out.s         : ratio of evidence distribution standard deviations assumed for the analysis.
    # out.type2_fit : output of fit_meta_d_MLE for the type 2 SDT fit.

    # 9/24/10 - bm - fixed program-crashing bug for (nR_S1, nR_S2) input
    # 9/7/10 - bm - wrote it

    ## parse inputs
    if len(stimID) == 0:
        raise ValueError("Empty data")

    if cellpadding is None:
        cellpadding = 1 / (2 * nRatings)

    # filters bad trials
    f = ((stimID == 0) | (stimID == 1)) & ((response == 0) | (response == 1)) & ((rating >= 1) & (rating <= nRatings))
    stimID = stimID[f]
    response = response[f]
    rating = rating[f]

    # convert to trial count format...
    nR_S1, nR_S2 = np.full(nRatings*2, np.nan), np.full(nRatings*2, np.nan)

    # get tallies of "S1" rating responses for S1 and S2 stim
    for i in range(nRatings):
        nR_S1[i] = np.sum((stimID == 0) & (response == 0) & (rating == nRatings - i))
        nR_S2[i] = np.sum((stimID == 1) & (response == 0) & (rating == nRatings - i))

    # get tallies of "S2" rating responses for S1 and S2 stim
    for i in range(nRatings):
        nR_S1[i + nRatings] = np.sum((stimID == 0) & (response == 1) & (rating == i + 1))
        nR_S2[i + nRatings] = np.sum((stimID == 1) & (response == 1) & (rating == i + 1))

    if np.any(nR_S1 == 0) | np.any(nR_S2 == 0):
        nR_S1 = nR_S1 + cellpadding
        nR_S2 = nR_S2 + cellpadding

    ## standard SDT analysis

    if equalVariance:
        s = 1

    ## type 2 SDT analysis

    fit = fit_meta_d_MLE(nR_S1, nR_S2, s)

    return fit


def type2roc(correct, conf, nbins=5):
    # Calculate area under type 2 ROC
    #
    # correct - vector of 1 x ntrials, 0 for error, 1 for correct
    # conf - vector of continuous confidence ratings between 0 and 1
    # nbins - how many bins to use for discretization

    bs = 1 / nbins
    H2, FA2 = np.full(nbins, np.nan), np.full(nbins, np.nan)
    for c in range(nbins):
        if c:
            H2[nbins - c - 1] = np.sum((conf > c*bs) & (conf <= (c+1)*bs) & (correct).astype(bool)) + 0.5
            FA2[nbins - c - 1] = np.sum((conf > c*bs) & (conf <= (c+1)*bs) & ~(correct).astype(bool)) + 0.5
        else:
            H2[nbins - c - 1] = np.sum((conf >= c * bs) & (conf <= (c + 1) * bs) & (correct).astype(bool)) + 0.5
            FA2[nbins - c - 1] = np.sum((conf >= c * bs) & (conf <= (c + 1) * bs) & ~(correct).astype(bool)) + 0.5

    H2 /= np.sum(H2)
    FA2 /= np.sum(FA2)
    cum_H2 = np.hstack((0, np.cumsum(H2)))
    cum_FA2 = np.hstack((0, np.cumsum(FA2)))

    k = np.full(nbins, np.nan)
    for c in range(nbins):
        k[c] = (cum_H2[c+1] - cum_FA2[c])**2 - (cum_H2[c] - cum_FA2[c+1])**2

    auroc2 = 0.5 + 0.25*np.sum(k)

    return auroc2

def type2roc_disc(correct, conf, nRatings=5):
    # Ratings have to start with 0!!!
    # Calculate area under type 2 ROC
    #
    # correct - vector of 1 x ntrials, 0 for error, 1 for correct
    # conf - vector of 1 x ntrials of confidence ratings taking values 0:Nratings-1
    # nRatings - how many confidence levels available

    correct = np.array(correct)
    conf = np.array(conf)

    H2, FA2 = np.full(nRatings, np.nan), np.full(nRatings, np.nan)
    for c in range(nRatings):
        H2[nRatings - c - 1] = np.sum((conf == c) & (correct).astype(bool)) + 0.5
        FA2[nRatings - c - 1] = np.sum((conf == c) & ~(correct).astype(bool)) + 0.5

    H2 /= np.sum(H2)
    FA2 /= np.sum(FA2)
    cum_H2 = np.hstack((0, np.cumsum(H2)))
    cum_FA2 = np.hstack((0, np.cumsum(FA2)))

    k = np.full(nRatings, np.nan)
    for c in range(nRatings):
        k[c] = (cum_H2[c+1] - cum_FA2[c])**2 - (cum_H2[c] - cum_FA2[c+1])**2

    auroc2 = 0.5 + 0.25*np.sum(k)

    return auroc2

def type2roc_disc2(stim, response, conf, nRatings=5):
    return type2roc_disc((np.array(stim) == np.array(response)).astype(int), conf, nRatings=nRatings)


def convert_for_mratio(stimuli, correct, confidence, nbins=3):
    #  see https://github.com/metacoglab/HMeta-d/wiki/HMeta-d-tutorial
    #  and http://www.columbia.edu/~bsm2105/type2sdt/archive/index.html

    bs = 1 / nbins
    nR_S = np.full((2, nbins * 2), np.nan)
    for s in range(2):  # iterate over presented stimuli
        for c in range(2):  # iterate over correctness
            conf = confidence[(stimuli == s) & (correct == c)]
            xrange = (range(nbins, 2*nbins), range(nbins-1, -1, -1))[(1, 0)[c] if s else c]
            for i in range(nbins):  # iterate over confidence bins
                if i == 0:
                    nR_S[s, xrange[i]] = np.sum((conf >= i * bs) & (conf <= (i + 1) * bs))
                else:
                    nR_S[s, xrange[i]] = np.sum((conf > i * bs) & (conf <= (i + 1) * bs))

    return nR_S[0], nR_S[1]


if __name__ == '__main__':
    nsamples = 10000
    stimulus = np.random.randint(0, 2, nsamples)
    choices = np.random.randint(0, 2, nsamples)
    rand_ind = np.random.permutation(range(nsamples))[:int(nsamples/2)]
    choices[rand_ind] = stimulus[rand_ind]
    confidence = 5 * np.ones(nsamples)
    print(type2_SDT_MLE(stimulus, choices, confidence, 5).M_ratio)
    from confidence.mca.type2_SDT_MLE_matlab import type2_SDT_MLE_matlab
    print(type2_SDT_MLE_matlab(stimulus, choices, confidence, 5).M_ratio)



import warnings

import numpy as np
from scipy.stats import lognorm, beta, betaprime, gamma

from .fast_truncnorm import truncnorm
from ._distributions import (TruncatedGumbelMode, TruncatedGumbelModeSD, TruncatedLognormal, TruncatedLognormalMode,
                             TruncatedLognormalMean, TruncatedLognormalModeSD, _params_lognormal_mode_std,
                             _params_truncnorm_mode_std)

TYPE2_NOISE_DISTS = (
    'beta_mode_std', 'beta_mode', 'beta_mean_std',
    'lognormal_mean_std', 'lognormal_mean', 'lognormal_mode_std', 'lognormal_mode', 'lognormal_median_std',
    'gamma_mean_std', 'gamma_mean', 'gamma_mean_cv', 'gamma_mode_std', 'gamma_mode',
    'betaprime_mean_std',
    'truncated_normal_mode_std', 'truncated_normal_mode',
    'truncated_gumbel_mode_std', 'truncated_gumbel_mode',
    'truncated_lognormal_mode_std', 'truncated_lognormal', 'truncated_lognormal_mode', 'truncated_lognormal_mean'
)

TYPE2_NOISE_DISTS_REPORT_ONLY = (
    'beta_mode_std', 'beta_mode', 'beta_mean_std',
    'truncated_lognormal_mode_std', 'truncated_lognormal_mode', 'truncated_lognormal', 'truncated_lognormal_mean'
)
TYPE2_NOISE_DISTS_READOUT_TEMPERATURE_ONLY = (
    'lognormal_mean_std', 'lognormal_mean', 'lognormal_mode_std', 'lognormal_mode', 'lognormal_median_std',
    'gamma_mean_std', 'gamma_mean', 'gamma_mean_cv', 'gamma_mode_std', 'gamma_mode',
    'betaprime_mean_std'
)



def get_type2_dist(type2_dist, type2_center, type2_noise, type2_noise_type='report'):
    """
    Helper function to select appropriately parameterized type 2 noise distributions.

    Parameters:
    -----------
    type2_dist : str
        Name of the type 2 noise distribution. Check TYPE2_NOISE_DISTS for possible values.
    type2_noise : float or array-like of dtype float
        "True" value of a noisy type 2 variable. Corresponds to type 1 noise (noisy-temperature), metacognitve
        evidence (noisy-readout) or confidence (noisy-report).
        For mean- and mode preserving distributions, the center corresponds to the mean (suffix _mean) and mode (suffix
        _mode), respectively.
    type2_noise : float or array-like of dtype float with mode.shape
        Spread parameter that represents metacognitve noise. For standard-deviation-preserving distributions, this
        parameter corresponds to the standard deviation when sampling data from the distribution (suffix _std).
    type2_noise_type : str (default='report')
        Metacognitive noise type. Possible values: 'report', 'readout', 'temperature'.

    Returns:
    --------
    scipy.stats continuous distribution instance
    """

    if type2_dist not in TYPE2_NOISE_DISTS:
        raise ValueError(f"Unkonwn distribution '{type2_dist}'.")
    elif (type2_noise_type == 'report') and type2_dist in TYPE2_NOISE_DISTS_READOUT_TEMPERATURE_ONLY:
        raise ValueError(f"Distribution '{type2_dist}' is only valid for noisy-readout or noisy-temperature models.")
    elif (type2_noise_type in ('readout', 'temperature')) and type2_dist in TYPE2_NOISE_DISTS_REPORT_ONLY:
        raise ValueError(f"Distribution '{type2_dist}' is only valid for noisy-report models.")

    if type2_noise < 1e-10:
        # warnings.warn('Type 2 noise is smaller than 1e-10, which can lead to unstable numerical results. It is set '
        #               'to a hard minimum of 1e-10.')
        type2_noise = 1e-10

    if type2_dist == 'lognormal_median_std':
        # type2_center = median, type2_noise = SD
        s = np.maximum(1e-12, np.sqrt(np.log((1 + np.sqrt(1 + 4 * (type2_noise / type2_center) ** 2)) / 2)))
        dist = lognorm(s=s, scale=type2_center)
    elif type2_dist == 'lognormal_mean':
        # type2_center = mean, type2_noise = SD in log space
        dist = lognorm(s=type2_noise, scale=np.exp(-type2_noise ** 2 / 2) * type2_center)
    elif type2_dist == 'lognormal_mode':
        # type2_center = mode, type2_noise = SD in log space
        if type2_noise > 23:
            # warnings.warn(f'Type 2 noise for a {type2_dist} distribution is too high. Capping at 23')
            type2_noise = 23
        dist = lognorm(s=type2_noise, scale=np.exp(type2_noise ** 2) * type2_center)
    elif type2_dist == 'lognormal_mode_std':
        # type2_center = mode, type2_noise = SD
        shape, type2_noise = _params_lognormal_mode_std(np.maximum(1e-5, type2_center), type2_noise)
        dist = lognorm(loc=0, scale=type2_noise, s=shape)
    elif type2_dist == 'lognormal_mean_std':
        # Corresponds to the CASSANDRE/LogN setup
        # type2_center = mean, type2_noise = SD
        type2_center = np.maximum(type2_center, 1e-12)
        sigma2 = np.log1p((np.maximum(type2_noise, 0) / type2_center) ** 2)
        scale = np.exp(np.log(type2_center) - sigma2 / 2)
        dist = lognorm(loc=0, scale=scale, s=np.sqrt(sigma2))
    elif type2_dist == 'beta_mean_std':
        # Canonical "precision" parameterization of the beta distribution, where precision = 1 / noise**2, i.e.
        # inverse variance.
        # type2_center = mean, type2_noise = SD

        type2_center = np.clip(type2_center, 1e-12, 1 - 1e-12)          # avoid a=0 or b=0
        type2_noise = np.maximum(type2_noise, 1e-12)

        phi = type2_center * (1 - type2_center) / (type2_noise ** 2) - 1
        phi = np.maximum(phi, 1e-12)

        a = type2_center * phi
        b = (1 - type2_center) * phi

        dist = beta(loc=0, a=a, b=b, scale=1)
    elif type2_dist == 'beta_mode_std':
        # type2_center = mode, type2_noise = SD

        type2_center = np.clip(type2_center, 1e-8, 1 - 1e-8)

        var_target = np.maximum(type2_noise ** 2, 1e-8)
        # Initial closed-form approximation
        kappa = type2_center * (1 - type2_center) / var_target
        # One-step variance correction
        num = type2_center * (1 - type2_center) * kappa ** 2 + kappa + 1
        den = (kappa + 2)**2 * (kappa + 3)
        var_actual = num / den
        kappa *= var_actual / var_target

        a = type2_center * kappa + 1
        b = (1 - type2_center) * kappa + 1
        dist = beta(loc=0, a=a, b=b, scale=1)

    elif type2_dist == 'beta_mode':
        # type2_center = mode, type2_noise != SD
        type2_center = np.maximum(1e-5, np.minimum(1 - 1e-5, type2_center))
        kappa = 2 + (type2_center * (1 - type2_center)) / (type2_noise ** 2)   # concentration/precision (>2)
        a = 1 + type2_center * (kappa - 2)
        b = 1 + (1 - type2_center) * (kappa - 2)
        dist = beta(loc=0, a=a, b=b, scale=1)
    elif type2_dist == 'betaprime_mean_std':
        # type2_center = mean, type2_noise = SD
        b = 2 + (type2_center * (type2_center + 1)) / type2_noise ** 2
        a = type2_center * (b - 1)
        dist = betaprime(loc=0, a=a, b=b, scale=1)
    elif type2_dist == 'gamma_mode_std':
        # type2_center = mode, type2_noise = SD
        u = (type2_center + np.sqrt(type2_center * type2_center + 4 * type2_noise ** 2)) / (2 * type2_noise)
        dist = gamma(loc=0, a=u**2, scale=type2_noise / u)
    elif type2_dist == 'gamma_mean_std':
        # type2_center = mean, type2_noise = SD
        dist = gamma(loc=0, a=(type2_center / type2_noise) ** 2, scale=(type2_noise ** 2) / type2_center)
    elif type2_dist == 'gamma_mean':
        # type2_center = mean, type2_noise != SD
        dist = gamma(loc=0, a=type2_center / type2_noise, scale=type2_noise)
    elif type2_dist == 'gamma_mean_cv':
        # Corresponds to the CASSANDRE setup, i.e. standard multiplicative noise
        # type2_center = center, type2_noise = CV (relative noise)
        a = 1 / (type2_noise ** 2)
        dist = gamma(loc=0, a=a, scale=type2_center / a)
    elif type2_dist == 'gamma_mode':
        # type2_center = mode; type2_noise != SD
        dist = gamma(loc=0, a=(type2_center / type2_noise) + 1, scale=type2_noise)
    elif type2_dist.startswith('truncated_'):
        if type2_noise_type == 'report':
            if type2_dist == 'truncated_normal_mode_std':
                # type2_center = mode, type2_noise = SD of the truncated distribution
                sigma = _params_truncnorm_mode_std(low=0, high=1, x=type2_center, noise=type2_noise)
                dist = truncnorm(a=-type2_center / sigma, b=(1 - type2_center) / sigma, loc=type2_center, scale=sigma)
            elif type2_dist == 'truncated_normal_mode':
                # type2_center = mode, type2_noise = SD of the *un*truncated distribution
                dist = truncnorm(a=-type2_center / type2_noise, b=(1 - type2_center) / type2_noise, loc=type2_center, scale=type2_noise)
            elif type2_dist == 'truncated_gumbel_mode_std':
                # type2_center = mode, type2_noise = SD of the truncated distribution
                dist = TruncatedGumbelModeSD(mode=type2_center, noise=type2_noise, upper=1, n_newton=3)
            elif type2_dist == 'truncated_gumbel_mode':
                # type2_center = mode, type2_noise = SD of the *un*truncated distribution
                dist = TruncatedGumbelMode(mode=type2_center, noise=type2_noise * np.sqrt(6) / np.pi, upper=1)
            elif type2_dist == 'truncated_lognormal_mode_std':
                # type2_center = mode, type2_noise = SD
                dist = TruncatedLognormalModeSD(mode=type2_center, noise=type2_noise, upper=1, n_newton=5)
            elif type2_dist == 'truncated_lognormal':
                # type2_center = median of the untruncated lognormal, type2_noise = SD in log space
                dist = TruncatedLognormal(median_untrunc=type2_center, noise=type2_noise, upper=1)
            elif type2_dist == 'truncated_lognormal_mode':
                # type2_center = mode, type2_noise = SD in log space
                if type2_noise > 25:
                    # warnings.warn(f'Type 2 noise for a {type2_dist} distribution is too high. Capping at 25')
                    type2_noise = 25
                dist = TruncatedLognormalMode(mode=type2_center, noise=type2_noise, b=1)
            elif type2_dist == 'truncated_lognormal_mean':
                # type2_center = mean, type2_noise = SD in log space
                dist = TruncatedLognormalMean(mean=type2_center, noise=type2_noise, upper=1)
        elif type2_noise_type in ('readout', 'temperature'):
            if type2_dist == 'truncated_normal_mode_std':
                # type2_center = mode, type2_noise = SD
                sigma = _params_truncnorm_mode_std(low=0, high=np.inf, x=type2_center, noise=type2_noise)
                dist = truncnorm(a=-type2_center / sigma, b=np.inf, loc=type2_center, scale=sigma)
            elif type2_dist == 'truncated_normal_mode':
                # type2_center = mode, type2_noise != SD
                dist = truncnorm(a=-type2_center / type2_noise, b=np.inf, loc=type2_center, scale=type2_noise)
            elif type2_dist == 'truncated_gumbel_mode_std':
                # type2_center = mode, type2_noise = SD
                dist = TruncatedGumbelModeSD(mode=type2_center, noise=type2_noise, upper=np.inf, n_newton=3)
            elif type2_dist == 'truncated_gumbel_mode':
                # type2_center = mode, type2_noise != SD
                dist = TruncatedGumbelMode(mode=type2_center, noise=type2_noise * np.sqrt(6) / np.pi, upper=np.inf)

        else:
            raise ValueError(f"'{type2_noise_type}' is an unknown type 2 noise type")

    return dist  # noqa


def get_likelihood(x, type2_dist, type2_center, type2_noise, binsize_meta=1e-3, logarithm=False):
    """
    Helper function to get the likelihood mass within type2_center ± binsize_meta for a type 2 noise distribution.
    """
    dist = get_type2_dist(type2_dist=type2_dist, type2_center=type2_center, type2_noise=type2_noise)
    likelihood = dist.cdf(x + binsize_meta) - dist.cdf(x - binsize_meta)
    return np.log(np.maximum(1e-4, likelihood)) if logarithm else likelihood